#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

######################################################################## Numba-accelerated interface

import awkward.array.base
import awkward.array.jagged
from .base import CppMethods
from .array_impl import JaggedArray

class JaggedArrayCpp(CppMethods, JaggedArray, awkward.array.jagged.JaggedArray):
    pass



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import operator
import math
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy
import numba
import numba.typing.arraydecl

import awkward.array.base
import awkward.array.jagged
from .base import NumbaMethods
from .base import AwkwardArrayType
from .base import clsrepr
from .base import ISADVANCED
from .base import NOTADVANCED
from .base import sliceval2
from .base import sliceval3

######################################################################## Numba-accelerated interface

class JaggedArrayNumba(NumbaMethods, awkward.array.jagged.JaggedArray):
    # @classmethod
    # def offsetsaliased(cls, starts, stops):
    ### base implementation is fine and don't need in Numba

    @classmethod
    def counts2offsets(cls, counts):
        return cls._counts2offsets(counts)

    @staticmethod
    @numba.njit
    def _counts2offsets(counts):
        offsets = numpy.empty(len(counts) + 1, dtype=numpy.int64)
        offsets[0] = 0
        for i in range(len(counts)):
            offsets[i + 1] = offsets[i] + counts[i]
        return offsets

    @classmethod
    def offsets2parents(cls, offsets):
        return cls._offsets2parents(offsets)

    @staticmethod
    @numba.njit
    def _offsets2parents(offsets):
        if len(offsets) == 0:
            raise ValueError("offsets must have at least one element")
        parents = numpy.empty(offsets[-1], dtype=numpy.int64)
        j = 0
        k = -1
        for i in offsets:
            while j < i:
                parents[j] = k
                j += 1
            k += 1
        return parents

    @classmethod
    def startsstops2parents(cls, starts, stops):
        return cls._startsstops2parents(starts, stops)

    @staticmethod
    @numba.njit
    def _startsstops2parents(starts, stops):
        out = numpy.full(stops.max(), -1, numpy.int64)
        for i in range(len(starts)):
            out[starts[i]:stops[i]] = i
        return out

    @classmethod
    def parents2startsstops(cls, parents, length=None):
        if length is None:
            length = parents.max() + 1
        return cls._parents2startsstops(parents, length)

    @staticmethod
    @numba.njit
    def _parents2startsstops(parents, length):
        starts = numpy.zeros(length, numpy.int64)
        stops = numpy.zeros(length, numpy.int64)

        last = -1
        for k in range(len(parents)):
            this = parents[k]
            if last != this:
                if last >= 0 and last < length:
                    stops[last] = k
                if this >= 0 and this < length:
                    starts[this] = k
            last = this

        if last != -1:
            stops[last] = len(parents)

        return starts, stops

    # @classmethod
    # def uniques2offsetsparents(cls, uniques):
    ### base implementation is fine and don't need in Numba

    # def __init__(self, starts, stops, content):
    ### base implementation is fine and already exposed in Numba

    @classmethod
    def fromiter(cls, iterable):
        import awkward.numba
        return awkward.numba.fromiter(iterable)

    # @classmethod
    # def fromoffsets(cls, offsets, content):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromcounts(cls, counts, content):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromparents(cls, parents, content, length=None):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromuniques(cls, uniques, content):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromindex(cls, index, content, validate=True):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromjagged(cls, jagged):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromregular(cls, regular):
    ### base implementation is fine and don't need in Numba

    # @classmethod
    # def fromfolding(cls, content, size):
    ### base implementation is fine and don't need in Numba

    # def copy(self, starts=None, stops=None, content=None):
    ### base implementation is fine and don't need in Numba

    # def deepcopy(self, starts=None, stops=None, content=None):
    ### base implementation is fine and don't need in Numba

    # def empty_like(self, **overrides):
    ### base implementation is fine and don't need in Numba

    # def zeros_like(self, **overrides):
    ### base implementation is fine and don't need in Numba

    # def ones_like(self, **overrides):
    ### base implementation is fine and don't need in Numba

    # def __awkward_persist__(self, ident, fill, prefix, suffix, schemasuffix, storage, compression, **kwargs):
    ### base implementation is fine and don't need in Numba

    # @property
    # def starts(self):
    ### base implementation is fine and already exposed in Numba

    # @starts.setter
    # def starts(self, value):
    ### base implementation is fine and don't need in Numba
        
    # @property
    # def stops(self):
    ### base implementation is fine and already exposed in Numba

    # @stops.setter
    # def stops(self, value):
    ### base implementation is fine and don't need in Numba

    # @property
    # def content(self):
    ### base implementation is fine and already exposed in Numba

    # @content.setter
    # def content(self, value):
    ### base implementation is fine and don't need in Numba

    # @property
    # def offsets(self):
    ### base implementation is fine and already exposed in Numba

    # @offsets.setter
    # def offsets(self, value):
    ### base implementation is fine and don't need in Numba

    # @property
    # def counts(self):
    ### base implementation is fine and already exposed in Numba

    # @counts.setter
    # def counts(self, value):
    ### base implementation is fine and don't need in Numba

    # @property
    # def parents(self):
    ### base implementation is fine and already exposed in Numba

    # @parents.setter
    # def parents(self, value):
    ### base implementation is fine and don't need in Numba

    # @property
    # def index(self):
    ### base implementation is fine and already exposed in Numba

    # def __len__(self):
    ### base implementation is fine and already exposed in Numba

    # def _gettype(self, seen):
    ### base implementation is fine and don't need in Numba

    def _valid(self):
        pass             # do validation in place from now on

    # @staticmethod
    # def _validstartsstops(starts, stops):
    ### base implementation is fine and don't need in Numba

    # def __iter__(self, checkiter=True):
    ### base implementation is fine and already exposed in Numba

    def __getitem__(self, where):
        if not isinstance(where, tuple):
            where = (where,)
        if len(where) == 0:
            return self

        newwhere = ()
        for x in where:
            if isinstance(x, Iterable) and not isinstance(x, (numpy.ndarray, awkward.array.base.AwkwardArray)):
                newwhere = newwhere + (numpy.array(x),)
            else:
                newwhere = newwhere + (x,)

        if len(newwhere) == 1:
            newwhere = newwhere[0]
            
        return self._getitem_impl(newwhere)

    @numba.njit
    def _getitem_impl(self, newwhere):
        return self[newwhere]

    # def __setitem__(self, where, what):

    @numba.generated_jit(nopython=True)
    def tojagged(self, data):
        assert not isinstance(data, JaggedArrayType)

        if isinstance(data, AwkwardArrayType):
            def impl(self, data):
                if len(self.starts) != len(data):
                    raise ValueError("cannot broadcast AwkwardArray to match JaggedArray with a different length")
                if len(self.starts.shape) != 1:
                    raise ValueError("cannot broadcast AwkwardArray to match JaggedArray that has len(starts.shape) != 1; call jagged.structure1d() first")
                index = numpy.empty(len(self.content), numpy.int64)
                for i in range(len(self.starts)):
                    index[self.starts[i]:self.stops[i]] = i
                return _JaggedArray_new(self, self.starts, self.stops, data[index], self.iscompact)
            return impl

        elif isinstance(data, numba.types.Array):
            def impl(self, data):
                if self.starts.shape != data.shape:
                    raise ValueError("cannot broadcast Numpy array to match a JaggedArray with a different length (or more generally, starts.shape)")
                content = numpy.empty(len(self.content), data.dtype)
                flatstarts = self.starts.reshape(-1)
                flatstops = self.stops.reshape(-1)
                flatdata = data.reshape(-1)
                for i in range(len(flatstarts)):
                    content[flatstarts[i]:flatstops[i]] = flatdata[i]
                return _JaggedArray_new(self, self.starts, self.stops, content, self.iscompact)
            return impl

        else:
            def impl(self, data):
                content = numpy.full(len(self.content), data)
                return _JaggedArray_new(self, self.starts, self.stops, content, self.iscompact)
            return impl

    # def _tojagged(self, starts=None, stops=None, copy=True):

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")
        if method != "__call__":
            return NotImplemented

        first = None
        inputs = list(inputs)
        for i in range(len(inputs)):
            if isinstance(inputs[i], awkward.array.jagged.JaggedArray):
                inputs[i] = inputs[i].compact()
                shift = inputs[i].starts[0]
                if shift != 0:
                    inputs[i] = inputs[i].copy(inputs[i].starts - shift, inputs[i].stops - shift, inputs[i].content[shift:])
                if first is None:
                    first = inputs[i]
                elif first.starts[0] != inputs[i].starts[0] or not numpy.array_equal(first.stops, inputs[i].stops):
                    raise ValueError("JaggedArrays in Numpy ufunc have incompatible structure")

        assert first is not None

        for i in range(len(inputs)):
            if isinstance(inputs[i], awkward.array.jagged.JaggedArray):
                pass
            elif isinstance(inputs[i], awkward.array.base.AwkwardArray):
                inputs[i] = first.tojagged(inputs[i])
            elif isinstance(inputs[i], Iterable):
                inputs[i] = first.tojagged(numpy.array(inputs[i], copy=False))
            else:
                inputs[i] = first.tojagged(inputs[i])

        for i in range(len(inputs)):
            inputs[i] = inputs[i]._content

        result = getattr(ufunc, method)(*inputs, **kwargs)

        if isinstance(result, tuple):
            return tuple(self.Methods.maybemixin(type(x), self.JaggedArray)(first.starts, first.stops, x) if isinstance(x, (numpy.ndarray, awkward.array.base.AwkwardArray)) else x for x in result)
        else:
            return self.Methods.maybemixin(type(result), self.JaggedArray)(first.starts, first.stops, result)

    @numba.njit
    def regular(self):
        return self.regular()

    ### FIXME: this whole section can't be done until we have Tables

    # def _argpairs(self):

    # def _argdistincts(self, absolute):

    # def argdistincts(self, nested=False):

    # def distincts(self, nested=False):

    # def argpairs(self, nested=False):

    # def pairs(self, nested=False):

    # def _argcross(self, other):

    # def argcross(self, other, nested=False):

    # def cross(self, other, nested=False):

    # def _canuseoffset(self):
    ### base implementation is fine and don't need in Numba

    # @property
    # def iscompact(self):
    ### base implementation is fine and already exposed in Numba

    @numba.njit
    def compact(self):
        return self.compact()

    def flatten(self, axis=0):
        if not self._util_isinteger(axis) or axis < 0:
            raise TypeError("axis must be a non-negative integer (can't count from the end)")
        if axis > 0:
            if isinstance(self._content, JaggedArray):
                counts = self.JaggedArray.fromcounts(self.counts, self._content.counts).sum()
                return self.JaggedArray.fromcounts(counts, self._content.flatten(axis=axis - 1))

        if len(self) == 0:
            return self._content[0:0]
        elif self.iscompact:
            return self._content[self._starts[0]:self.stops[-1]]  # no underscore in stops
        else:
            out = self.compact()
            return out._content[out._starts[0]:out.stops[0]]      # no underscore in stops

    # def structure1d(self, levellimit=None):
    ### base implementation is fine and can't(?) be exposed in Numba (type manipulation is hard!)

    # def _hasjagged(self):
    ### base implementation is fine

    # def _reduce(self, ufunc, identity, dtype):

    @numba.njit
    def argmin(self):
        return self.argmin()

    @numba.njit
    def argmax(self):
        return self.argmax()

    def _argminmax(self, ismin):
        raise RuntimeError("helper function not needed in JaggedArrayNumba")

    def _argminmax_general(self, ismin):
        raise RuntimeError("helper function not needed in JaggedArrayNumba")

    # def _concatenate_axis0(isclassmethod, cls_or_self, arrays):
    #     if isinstance(arrays, (numpy.ndarray, awkward.array.base.AwkwardArray)):
    #         arrays = (arrays,)
    #     else:
    #         arrays = tuple(arrays)
    #     if isclassmethod:
    #         cls = cls_or_self
    #         if not all(isinstance(x, awkward.array.jagged.JaggedArray) for x in arrays):
    #             raise TypeError("cannot concatenate non-JaggedArrays with JaggedArray.concatenate")
    #     else:
    #         self = cls_or_self
    #         cls = self.__class__
    #         if not isinstance(self, awkward.array.jagged.JaggedArray) or not all(isinstance(x, awkward.array.jagged.JaggedArray) for x in arrays):
    #             raise TypeError("cannot concatenate non-JaggedArrays with JaggedArray.concatenate")
    #         arrays = (self,) + arrays
    #     if len(arrays) == 0:
    #         raise TypeError("concatenate requires at least one array")
    #     return _JaggedArray_concatenate_njit(arrays, axis)
    ### FIXME: left unfinished

    # @awkward.util.bothmethod
    # def zip(isclassmethod, cls_or_self, columns1={}, *columns2, **columns3):
    ### FIXME: can't do this one until we have Tables

    # def pad(self, length, maskedwhen=True, clip=False):

######################################################################## register types in Numba

@numba.extending.typeof_impl.register(awkward.array.jagged.JaggedArray)
def _JaggedArray_typeof(val, c):
    return JaggedArrayType(numba.typeof(val.starts), numba.typeof(val.stops), numba.typeof(val.content), special=type(val))

class JaggedArrayType(AwkwardArrayType):
    def __init__(self, startstype, stopstype, contenttype, special=awkward.array.jagged.JaggedArray):
        super(JaggedArrayType, self).__init__(name="JaggedArrayType({0}, {1}, {2}{3})".format(startstype.name, stopstype.name, contenttype.name, "" if special is awkward.array.jagged.JaggedArray else clsrepr(special)))
        if startstype.ndim != stopstype.ndim:
            raise ValueError("JaggedArray.starts must have the same number of dimensions as JaggedArray.stops")
        if startstype.ndim == 0:
            raise ValueError("JaggedArray.starts and JaggedArray.stops must have at least one dimension")
        self.startstype = startstype
        self.stopstype = stopstype
        self.contenttype = contenttype
        self.special = special

    def getitem(self, wheretype):
        if self.startstype.ndim > 1 and not any(isinstance(x, (numba.types.Array, JaggedArrayType)) for x in wheretype.types[:self.startstype.ndim]):
            headtype = numba.types.Tuple(wheretype.types[:self.startstype.ndim])
            tailtype = numba.types.Tuple(wheretype.types[self.startstype.ndim:])

            outstartstype = numba.typing.arraydecl.get_array_index_type(self.startstype, headtype).result
            outstopstype = numba.typing.arraydecl.get_array_index_type(self.stopstype, headtype).result
            if isinstance(self.contenttype, JaggedArrayType):
                outcontenttype = self.contenttype.getitem(tailtype)
            else:
                outcontenttype = numba.typing.arraydecl.get_array_index_type(self.contenttype, tailtype).result

            assert isinstance(outstartstype, numba.types.Array) == isinstance(outstopstype, numba.types.Array)
            if isinstance(outstartstype, numba.types.Array):
                return JaggedArrayType(outstartstype, outstopstype, outcontenttype, special=self.special)
            else:
                return outcontenttype

        else:
            headtype = wheretype.types[0]
            tailtype = numba.types.Tuple(wheretype.types[1:])

            if isinstance(headtype, JaggedArrayType) and len(tailtype.types) == 0:
                return _JaggedArray_typer_getitem_jagged(self, headtype)

            else:
                fake = _JaggedArray_typer_getitem(JaggedArrayType(JaggedArrayNumba.NUMBA_INDEXTYPE[:], JaggedArrayNumba.NUMBA_INDEXTYPE[:], self), wheretype, NOTADVANCED)
                if isinstance(fake, numba.types.Array):
                    return fake.dtype
                else:
                    return fake.contenttype

    @property
    def len_impl(self):
        return _JaggedArray_lower_len

    @property
    def getitem_impl(self):
        return lambda context, builder, sig, args: _JaggedArray_lower_getitem_integer(context, builder, sig, args, checkvalid=False)

def _JaggedArray_typer_getitem_jagged(arraytype, headtype):
    if isinstance(headtype.contenttype, JaggedArrayType):
        if not isinstance(arraytype.contenttype, JaggedArrayType):
            raise TypeError("index (in square brackets) is more deeply jagged than array (before square brackets)")
        contenttype = _JaggedArray_typer_getitem_jagged(arraytype.contenttype, headtype.contenttype)
    elif isinstance(headtype.contenttype, numba.types.Array) and headtype.contenttype.ndim == 1 and isinstance(headtype.contenttype.dtype, (numba.types.Boolean, numba.types.Integer)):
        contenttype = arraytype.contenttype
    else:
        raise TypeError("jagged indexing must be boolean or integers with 1-dimensional content")

    return JaggedArrayType(arraytype.startstype, arraytype.stopstype, contenttype, special=arraytype.special)

def _JaggedArray_typer_getitem(arraytype, wheretype, advancedtype):
    if len(wheretype.types) == 0:
        return arraytype

    if arraytype.startstype.ndim != 1 or arraytype.stopstype.ndim != 1:
        raise NotImplementedError("multidimensional starts and stops not supported; call jagged.structure1d() first")

    isarray = (isinstance(wheretype.types[0], numba.types.Array) and wheretype.types[0].ndim == 1)

    contenttype = _JaggedArray_typer_getitem(arraytype.contenttype, numba.types.Tuple(wheretype.types[1:]), ISADVANCED if isarray else advancedtype)

    if isinstance(wheretype.types[0], numba.types.Integer) or (isarray and advancedtype == ISADVANCED):
        return contenttype
    elif isinstance(wheretype.types[0], numba.types.SliceType) or (isarray and advancedtype == NOTADVANCED):
        return JaggedArrayType(arraytype.startstype, arraytype.stopstype, contenttype, special=arraytype.special)
    elif isinstance(wheretype.types[0], JaggedArrayType):
        raise TypeError("cannot use jagged indexing in a tuple")
    else:
        raise TypeError("cannot be used for indexing: {0}".format(wheretype.types[0]))

def _JaggedArray_getitem_next(array, where):
    pass

@numba.extending.type_callable(_JaggedArray_getitem_next)
def _JaggedArray_type_getitem_next(context):
    return _JaggedArray_typer_getitem

######################################################################## model and boxing

@numba.extending.register_model(JaggedArrayType)
class JaggedArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("starts", fe_type.startstype),
                   ("stops", fe_type.stopstype),
                   ("content", fe_type.contenttype),
                   ("iscompact", JaggedArrayNumba.NUMBA_BOOLTYPE)]
        super(JaggedArrayModel, self).__init__(dmm, fe_type, members)

numba.extending.make_attribute_wrapper(JaggedArrayType, "starts", "starts")
numba.extending.make_attribute_wrapper(JaggedArrayType, "stops", "stops")
numba.extending.make_attribute_wrapper(JaggedArrayType, "content", "content")
numba.extending.make_attribute_wrapper(JaggedArrayType, "iscompact", "iscompact")

@numba.extending.unbox(JaggedArrayType)
def _JaggedArray_unbox(typ, obj, c):
    starts_obj = c.pyapi.object_getattr_string(obj, "starts")
    stops_obj = c.pyapi.object_getattr_string(obj, "stops")
    content_obj = c.pyapi.object_getattr_string(obj, "content")
    iscompact_obj = c.pyapi.object_getattr_string(obj, "iscompact")

    array = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder)
    array.starts = c.pyapi.to_native_value(typ.startstype, starts_obj).value
    array.stops = c.pyapi.to_native_value(typ.stopstype, stops_obj).value
    array.content = c.pyapi.to_native_value(typ.contenttype, content_obj).value
    array.iscompact = c.pyapi.to_native_value(JaggedArrayNumba.NUMBA_BOOLTYPE, iscompact_obj).value

    c.pyapi.decref(starts_obj)
    c.pyapi.decref(stops_obj)
    c.pyapi.decref(content_obj)
    c.pyapi.decref(iscompact_obj)

    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(array._getvalue(), is_error)

@numba.extending.box(JaggedArrayType)
def _JaggedArray_box(typ, val, c):
    array = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    starts_obj = c.pyapi.from_native_value(typ.startstype, array.starts, c.env_manager)
    stops_obj = c.pyapi.from_native_value(typ.stopstype, array.stops, c.env_manager)
    content_obj = c.pyapi.from_native_value(typ.contenttype, array.content, c.env_manager)

    cls = c.pyapi.unserialize(c.pyapi.serialize_object(typ.special))
    out = c.pyapi.call_function_objargs(cls, (starts_obj, stops_obj, content_obj))

    c.pyapi.decref(starts_obj)
    c.pyapi.decref(stops_obj)
    c.pyapi.decref(content_obj)

    return out

@numba.extending.type_callable(awkward.array.jagged.JaggedArray)
def _JaggedArray_type_init(context):
    def typer(startstype, stopstype, contenttype):
        if isinstance(startstype, numba.types.Array) and isinstance(stopstype, numba.types.Array) and isinstance(contenttype, (numba.types.Array, AwkwardArrayType)):
            return JaggedArrayType(startstype, stopstype, contenttype, special=awkward.array.jagged.JaggedArray)
    return typer

@numba.extending.type_callable(JaggedArrayNumba)
def _JaggedArray_type_init(context):
    def typer(startstype, stopstype, contenttype):
        if isinstance(startstype, numba.types.Array) and isinstance(stopstype, numba.types.Array) and isinstance(contenttype, (numba.types.Array, AwkwardArrayType)):
            return JaggedArrayType(startstype, stopstype, contenttype, special=JaggedArrayNumba)
    return typer

@numba.extending.lower_builtin(awkward.array.jagged.JaggedArray, numba.types.Array, numba.types.Array, numba.types.Array)
@numba.extending.lower_builtin(awkward.array.jagged.JaggedArray, numba.types.Array, numba.types.Array, AwkwardArrayType)
@numba.extending.lower_builtin(JaggedArrayNumba, numba.types.Array, numba.types.Array, numba.types.Array)
@numba.extending.lower_builtin(JaggedArrayNumba, numba.types.Array, numba.types.Array, AwkwardArrayType)
def _JaggedArray_init_array(context, builder, sig, args):
    startstype, stopstype, contenttype = sig.args
    startsval, stopsval, contentval = args

    if context.enable_nrt:
        context.nrt.incref(builder, startstype, startsval)
        context.nrt.incref(builder, stopstype, stopsval)
        context.nrt.incref(builder, contenttype, contentval)

    array = numba.cgutils.create_struct_proxy(sig.return_type)(context, builder)
    array.starts = startsval
    array.stops = stopsval
    array.content = contentval
    array.iscompact = context.get_constant(JaggedArrayNumba.NUMBA_BOOLTYPE, False)   # unless you reproduce that logic here or call out to Python
    return array._getvalue()

def _JaggedArray_new(array, starts, stops, content, iscompact):
    pass

@numba.extending.type_callable(_JaggedArray_new)
def _JaggedArray_type_new(context):
    def typer(arraytype, startstype, stopstype, contenttype, iscompacttype):
        return JaggedArrayType(startstype, stopstype, contenttype, special=arraytype.special)
    return typer

@numba.extending.lower_builtin(_JaggedArray_new, JaggedArrayType, numba.types.Array, numba.types.Array, numba.types.Array, numba.types.Boolean)
@numba.extending.lower_builtin(_JaggedArray_new, JaggedArrayType, numba.types.Array, numba.types.Array, AwkwardArrayType, numba.types.Boolean)
def _JaggedArray_lower_new(context, builder, sig, args):
    arraytype, startstype, stopstype, contenttype, iscompacttype = sig.args
    arrayval, startsval, stopsval, contentval, iscompactval = args

    if context.enable_nrt:
        context.nrt.incref(builder, startstype, startsval)
        context.nrt.incref(builder, stopstype, stopsval)
        context.nrt.incref(builder, contenttype, contentval)

    array = numba.cgutils.create_struct_proxy(sig.return_type)(context, builder)
    array.starts = startsval
    array.stops = stopsval
    array.content = contentval
    array.iscompact = iscompactval
    return array._getvalue()

######################################################################## utilities

def _check_startstop_contentlen(context, builder, starttype, startval, stoptype, stopval, contenttype, contentval):
    if isinstance(contenttype, numba.types.Array):
        contentlen = numba.targets.arrayobj.array_len(context, builder, numba.types.intp(contenttype), (contentval,))
    else:
        contentlen = _JaggedArray_lower_len(context, builder, numba.types.intp(contenttype), (contentval,))

    with builder.if_then(builder.or_(builder.or_(builder.icmp_signed("<", startval, context.get_constant(starttype, 0)),
                                                 builder.icmp_signed("<", stopval, context.get_constant(stoptype, 0))),
                                     builder.or_(builder.icmp_signed(">=", startval, contentlen),
                                                 builder.icmp_signed(">", stopval, contentlen))),
                         likely=False):
        context.call_conv.return_user_exc(builder, ValueError, ("JaggedArray.starts or JaggedArray.stops is beyond the range of JaggedArray.content",))

######################################################################## lowered len

@numba.extending.lower_builtin(len, JaggedArrayType)
def _JaggedArray_lower_len(context, builder, sig, args):
    arraytype, = sig.args
    arrayval, = args

    array = numba.cgutils.create_struct_proxy(arraytype)(context, builder, value=arrayval)

    return numba.targets.arrayobj.array_len(context, builder, numba.types.intp(arraytype.startstype), (array.starts,))

######################################################################## lowered getitem

@numba.extending.lower_builtin(operator.getitem, JaggedArrayType, numba.types.Integer)
def _JaggedArray_lower_getitem_integer(context, builder, sig, args, checkvalid=True):
    arraytype, wheretype = sig.args
    arrayval, whereval = args

    array = numba.cgutils.create_struct_proxy(arraytype)(context, builder, value=arrayval)

    startstype = arraytype.startstype
    stopstype = arraytype.stopstype
    contenttype = arraytype.contenttype

    if startstype.ndim == 1:
        start = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, startstype.dtype(startstype, wheretype), (array.starts, whereval))
        stop = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, stopstype.dtype(stopstype, wheretype), (array.stops, whereval))

        if checkvalid:
            _check_startstop_contentlen(context, builder, startstype.dtype, start, stopstype.dtype, stop, contenttype, array.content)

        if isinstance(contenttype, numba.types.Array):
            return numba.targets.arrayobj.getitem_arraynd_intp(context, builder, contenttype(contenttype, numba.types.slice2_type), (array.content, sliceval2(context, builder, start, stop)))
        else:
            return _JaggedArray_lower_getitem_slice(context, builder, contenttype(contenttype, numba.types.slice2_type), (array.content, sliceval2(context, builder, start, stop)))

    else:
        outstartstype = numba.types.Array(startstype.dtype, startstype.ndim - 1, startstype.layout)
        outstopstype = numba.types.Array(stopstype.dtype, stopstype.ndim - 1, stopstype.layout)

        starts = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, outstartstype(startstype, wheretype), (array.starts, whereval))
        stops = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, outstopstype(stopstype, wheretype), (array.stops, whereval))

        outtype = JaggedArrayType(outstartstype, outstopstype, contenttype, special=arraytype.special)
        return _JaggedArray_lower_new(context, builder, outtype(arraytype, outstartstype, outstopstype, contenttype, JaggedArrayNumba.NUMBA_BOOLTYPE), (arrayval, starts, stops, array.content, array.iscompact))

@numba.extending.lower_builtin(operator.getitem, JaggedArrayType, numba.types.SliceType)
def _JaggedArray_lower_getitem_slice(context, builder, sig, args):
    arraytype, wheretype = sig.args
    arrayval, whereval = args

    array = numba.cgutils.create_struct_proxy(arraytype)(context, builder, value=arrayval)

    startstype = arraytype.startstype
    starts = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, startstype(startstype, wheretype), (array.starts, whereval))

    stopstype = arraytype.stopstype
    stops = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, stopstype(stopstype, wheretype), (array.stops, whereval))

    slice = context.make_helper(builder, wheretype, value=whereval)
    iscompact = builder.and_(array.iscompact,
                             builder.or_(builder.icmp_signed("==", slice.step, context.get_constant(numba.types.intp, 1)),
                                         builder.icmp_signed("==", slice.step, context.get_constant(numba.types.intp, numba.types.intp.maxval))))

    contenttype = arraytype.contenttype
    return _JaggedArray_lower_new(context, builder, arraytype(arraytype, startstype, stopstype, contenttype, JaggedArrayNumba.NUMBA_BOOLTYPE), (arrayval, starts, stops, array.content, iscompact))

@numba.extending.lower_builtin(operator.getitem, JaggedArrayType, numba.types.Array)
def _JaggedArray_lower_getitem_array(context, builder, sig, args):
    arraytype, wheretype = sig.args
    arrayval, whereval = args

    array = numba.cgutils.create_struct_proxy(arraytype)(context, builder, value=arrayval)

    startstype = arraytype.startstype
    starts = numba.targets.arrayobj.fancy_getitem_array(context, builder, startstype(startstype, wheretype), (array.starts, whereval))

    stopstype = arraytype.stopstype
    stops = numba.targets.arrayobj.fancy_getitem_array(context, builder, stopstype(stopstype, wheretype), (array.stops, whereval))

    contenttype = arraytype.contenttype
    return _JaggedArray_lower_new(context, builder, arraytype(arraytype, startstype, stopstype, contenttype, JaggedArrayNumba.NUMBA_BOOLTYPE), (arrayval, starts, stops, array.content, context.get_constant(JaggedArrayNumba.NUMBA_BOOLTYPE, False)))

@numba.extending.lower_builtin(operator.getitem, JaggedArrayType, JaggedArrayType)
def _JaggedArray_lower_getitem_jaggedarray(context, builder, sig, args):
    arraytype, wheretype = sig.args

    if isinstance(wheretype, JaggedArrayType) and isinstance(wheretype.contenttype, JaggedArrayType):
        def getitem(array, where):
            return _JaggedArray_new(array, array.starts, array.stops, array.content[where.content], True)

    elif isinstance(wheretype, JaggedArrayType) and isinstance(wheretype.contenttype.dtype, numba.types.Boolean) and isinstance(arraytype.contenttype, numba.types.Array):
        def getitem(array, where):
            if len(array) != len(where):
                raise IndexError("jagged index must have the same (outer) length as the JaggedArray it indexes")
            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            offsets[0] = 0
            content = numpy.empty(where.content.astype(numpy.int64).sum(), array.content.dtype)
            k = 0
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                wherei = where[i]
                if len(wherei) > length:
                    raise IndexError("jagged index is out of bounds in JaggedArray")

                for j in range(len(wherei)):
                    if wherei[j]:
                        content[k] = array.content[array.starts[i] + j]
                        k += 1
                offsets[i + 1] = k

            starts = offsets[:-1]
            stops = offsets[1:]
            return _JaggedArray_new(array, starts, stops, content, True)

    elif isinstance(wheretype, JaggedArrayType) and isinstance(wheretype.contenttype.dtype, numba.types.Boolean):
        def getitem(array, where):
            if len(array) != len(where):
                raise IndexError("jagged index must have the same (outer) length as the JaggedArray it indexes")
            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            offsets[0] = 0
            index = numpy.empty(len(where.content), numpy.int64)
            k = 0
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                wherei = where[i]
                if len(wherei) > length:
                    raise IndexError("jagged index is out of bounds in JaggedArray")

                for j in range(len(wherei)):
                    if wherei[j]:
                        index[k] = array.starts[i] + j
                        k += 1
                offsets[i + 1] = k

            starts = offsets[:-1]
            stops = offsets[1:]
            return _JaggedArray_new(array, starts, stops, array.content[index[:k]], True)

    elif isinstance(wheretype, JaggedArrayType) and isinstance(wheretype.contenttype.dtype, numba.types.Integer) and isinstance(arraytype.contenttype, numba.types.Array):
        def getitem(array, where):
            if len(array) != len(where):
                raise IndexError("jagged index must have the same (outer) length as the JaggedArray it indexes")
            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            offsets[0] = 0
            content = numpy.empty(len(where.content), array.content.dtype)
            k = 0
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                wherei = where[i]

                for j in range(len(wherei)):
                    norm = wherei[j]
                    if norm < 0:
                        norm += length
                    if norm < 0 or norm >= length:
                        raise IndexError("jagged index is out of bounds in JaggedArray")
                    content[k] = array.content[array.starts[i] + norm]
                    k += 1
                offsets[i + 1] = k

            starts = offsets[:-1]
            stops = offsets[1:]
            return _JaggedArray_new(array, starts, stops, content, True)

    elif isinstance(wheretype, JaggedArrayType) and isinstance(wheretype.contenttype.dtype, numba.types.Integer):
        def getitem(array, where):
            if len(array) != len(where):
                raise IndexError("jagged index must have the same (outer) length as the JaggedArray it indexes")
            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            offsets[0] = 0
            index = numpy.empty(len(where.content), numpy.int64)
            k = 0
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                wherei = where[i]

                for j in range(len(wherei)):
                    norm = wherei[j]
                    if norm < 0:
                        norm += length
                    if norm < 0 or norm >= length:
                        raise IndexError("jagged index is out of bounds in JaggedArray")
                    index[k] = array.starts[i] + norm
                    k += 1
                offsets[i + 1] = k

            starts = offsets[:-1]
            stops = offsets[1:]
            return _JaggedArray_new(array, starts, stops, array.content[index], True)

    else:
        raise AssertionError(where)

    return context.compile_internal(builder, getitem, sig, args)

@numba.extending.lower_builtin(operator.getitem, JaggedArrayType, numba.types.BaseTuple)
def _JaggedArray_lower_getitem_enter(context, builder, sig, args):
    arraytype, wheretype = sig.args
    arrayval, whereval = args

    if len(wheretype.types) == 1:
        if isinstance(wheretype.types[0], numba.types.Integer):
            getitem = _JaggedArray_lower_getitem_integer
        elif isinstance(wheretype.types[0], numba.types.SliceType):
            getitem = _JaggedArray_lower_getitem_slice
        elif isinstance(wheretype.types[0], numba.types.Array):
            getitem = _JaggedArray_lower_getitem_array
        return getitem(context, builder, sig.return_type(arraytype, wheretype.types[0]), (arrayval, builder.extract_value(whereval, 0)))

    if any(isinstance(x, numba.types.Array) for x in wheretype.types):
        arraylen = numba.cgutils.alloca_once_value(builder, context.get_constant(JaggedArrayNumba.NUMBA_INDEXTYPE, 0))
        for i, whereitemtype in enumerate(wheretype.types):
            if isinstance(whereitemtype, numba.types.Array):
                if isinstance(whereitemtype.dtype, numba.types.Boolean):
                    enter_arraylen = lambda whereitem, arraylen: max(arraylen, whereitem.astype(numpy.int64).sum())
                else:
                    enter_arraylen = lambda whereitem, arraylen: max(arraylen, len(whereitem))

                whereitemval = builder.extract_value(whereval, i)
                arraylenval = context.compile_internal(builder, enter_arraylen, JaggedArrayNumba.NUMBA_INDEXTYPE(whereitemtype, JaggedArrayNumba.NUMBA_INDEXTYPE), (whereitemval, builder.load(arraylen)))
                builder.store(arraylenval, arraylen)

        arraylenval = builder.load(arraylen)
        newwheretype = []
        newwherevals = []
        for i, old in enumerate(wheretype.types):
            if isinstance(old, numba.types.Array) and isinstance(old.dtype, numba.types.Boolean):
                toadvanced = lambda whereitem, arraylen: numpy.where(whereitem)[0]
            elif isinstance(old, numba.types.Array):
                toadvanced = lambda whereitem, arraylen: numpy.full(arraylen, whereitem[0], numpy.int64) if len(whereitem) == 1 else whereitem
            elif isinstance(old, numba.types.Integer):
                toadvanced = lambda whereitem, arraylen: numpy.full(arraylen, whereitem, numpy.int64)
            else:
                toadvanced = None

            whereitemval = builder.extract_value(whereval, i)
            if toadvanced is None:
                newwheretype.append(old)
                newwherevals.append(whereitemval)
            else:
                new = numba.types.Array(JaggedArrayNumba.NUMBA_INDEXTYPE, 1, "C") if isinstance(old, (numba.types.Array, numba.types.Integer)) else old
                newwheretype.append(new)
                newwherevals.append(context.compile_internal(builder, toadvanced, new(old, JaggedArrayNumba.NUMBA_INDEXTYPE), (whereitemval, arraylenval)))

        wheretype = numba.types.Tuple(tuple(newwheretype))
        whereval = context.make_tuple(builder, wheretype, tuple(newwherevals))

    def fake1(array, where):
        return _JaggedArray_getitem_next(awkward.array.jagged.JaggedArray(numpy.array([0], numpy.int64), numpy.array([len(array)], numpy.int64), array), where, None)[0]

    def fake2(array, where):
        out = _JaggedArray_getitem_next(awkward.array.jagged.JaggedArray(numpy.array([0], numpy.int64), numpy.array([len(array)], numpy.int64), array), where, None)
        return out.content[out.starts[0]:out.stops[-1]]

    fake = fake1 if all(isinstance(x, numba.types.Integer) for x in wheretype.types) else fake2
    return context.compile_internal(builder, fake, sig.return_type(arraytype, wheretype), (arrayval, whereval))

@numba.generated_jit(nopython=True)
def _JaggedArray_getitem_enter_toadvanced(whereitem, arraylen):
    if isinstance(whereitem, numba.types.Array) and isinstance(whereitem.dtype, numba.types.Boolean):
        return lambda whereitem, arraylen: numpy.nonzero(whereitem)[0]
    elif isinstance(whereitem, numba.types.Array):
        return lambda whereitem, arraylen: numpy.full(arraylen, whereitem[0], numpy.int64) if len(whereitem) == 1 else whereitem
    elif isinstance(whereitem, numba.types.Integer):
        return lambda whereitem, arraylen: numpy.full(arraylen, whereitem, numpy.int64)
    else:
        return lambda whereitem, arraylen: whereitem

@numba.extending.lower_builtin(_JaggedArray_getitem_next, numba.types.Array, numba.types.BaseTuple, numba.types.NoneType)
@numba.extending.lower_builtin(_JaggedArray_getitem_next, JaggedArrayType, numba.types.BaseTuple, numba.types.NoneType)
@numba.extending.lower_builtin(_JaggedArray_getitem_next, numba.types.Array, numba.types.BaseTuple, numba.types.Array)
@numba.extending.lower_builtin(_JaggedArray_getitem_next, JaggedArrayType, numba.types.BaseTuple, numba.types.Array)
def _JaggedArray_lower_getitem_next(context, builder, sig, args):
    arraytype, wheretype, advancedtype = sig.args
    arrayval, whereval, advancedval = args

    if len(wheretype.types) == 0:
        if context.enable_nrt:
            context.nrt.incref(builder, arraytype, arrayval)
        return arrayval

    headtype = wheretype.types[0]
    tailtype = numba.types.Tuple(wheretype.types[1:])
    headval = numba.targets.tupleobj.static_getitem_tuple(context, builder, headtype(wheretype, JaggedArrayNumba.NUMBA_INDEXTYPE), (whereval, 0))
    tailval = numba.targets.tupleobj.static_getitem_tuple(context, builder, tailtype(wheretype, numba.types.slice2_type), (whereval, slice(1, None)))

    if isinstance(headtype, numba.types.Integer):
        if isinstance(arraytype.contenttype, numba.types.Array):
            def getitem(array, head, tail, advanced):
                content = numpy.empty(len(array.starts), array.content.dtype)
                for i in range(len(array.starts)):
                    norm = head
                    if norm < 0:
                        norm += array.stops[i] - array.starts[i]
                    j = array.starts[i] + norm
                    if j >= array.stops[i]:
                        raise ValueError("integer index is beyond the range of one of the JaggedArray starts/stops pairs")
                    content[i] = array.content[j]
                return _JaggedArray_getitem_next(content, tail, advanced)

        else:
            def getitem(array, head, tail, advanced):
                index = numpy.empty(len(array.starts), numpy.int64)
                for i in range(len(array.starts)):
                    norm = head
                    if norm < 0:
                        norm += array.stops[i] - array.starts[i]
                    j = array.starts[i] + norm
                    if j >= array.stops[i]:
                        raise ValueError("integer index is beyond the range of one of the JaggedArray starts/stops pairs")
                    index[i] = j
                return _JaggedArray_getitem_next(array.content[index], tail, advanced)

    elif isinstance(headtype, numba.types.SliceType) and headtype.members == 2 and advancedtype == NOTADVANCED and not any(isinstance(x, numba.types.Array) for x in tailtype):
        intp_maxval = numba.types.intp.maxval

        def getitem(array, head, tail, advanced):
            if (head.start == 0 or head.start == intp_maxval) and head.stop == intp_maxval:
                next = _JaggedArray_getitem_next(array.content, tail, advanced)
                return _JaggedArray_new(array, array.starts, array.stops, next, array.iscompact)

            starts = numpy.empty(len(array.starts), numpy.int64)
            stops = numpy.empty(len(array.starts), numpy.int64)
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                a = head.start
                b = head.stop

                if a == intp_maxval:
                    a = 0
                elif a < 0:
                    a += length
                if b == intp_maxval:
                    b = length
                elif b < 0:
                    b += length

                if b <= a:
                    a = 0
                    b = 0
                if a < 0:
                    a = 0
                elif a > length:
                    a = length
                if b < 0:
                    b = 0
                elif b > length:
                    b = length

                starts[i] = array.starts[i] + a
                stops[i] = array.starts[i] + b

            next = _JaggedArray_getitem_next(array.content, tail, advanced)
            return _JaggedArray_new(array, starts, stops, next, False)

    elif isinstance(headtype, numba.types.SliceType):
        intp_maxval = numba.types.intp.maxval

        def getitem(array, head, tail, advanced):
            if head.step == 0:
                raise ValueError("slice step cannot be zero")

            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            offsets[0] = 0
            index = numpy.empty(len(array.content), numpy.int64)
            k = 0
            for i in range(len(array.starts)):
                length = array.stops[i] - array.starts[i]
                a = head.start
                b = head.stop
                c = head.step
                if c == intp_maxval:
                    c = 1

                if a == intp_maxval and c > 0:
                    a = 0
                elif a == intp_maxval:
                    a = length - 1
                elif a < 0:
                    a += length

                if b == intp_maxval and c > 0:
                    b = length
                elif b == intp_maxval:
                    b = -1
                elif b < 0:
                    b += length

                if c > 0:
                    if b <= a:
                        a = 0
                        b = 0
                    if a < 0:
                        a = 0
                    elif a > length:
                        a = length
                    if b < 0:
                        b = 0
                    elif b > length:
                        b = length
                else:
                    if a <= b:
                        a = 0
                        b = 0
                    if a < -1:
                        a = -1
                    elif a >= length:
                        a = length - 1
                    if b < -1:
                        b = -1
                    elif b >= length:
                        b = length - 1

                for j in range(a, b, c):
                    index[k] = array.starts[i] + j
                    k += 1
                offsets[i + 1] = k

            starts = offsets[:-1]
            stops = offsets[1:]
            next = _JaggedArray_getitem_next(array.content[index[:k]], tail, _spread_advanced(starts, stops, advanced))
            return _JaggedArray_new(array, starts, stops, next, True)

    elif isinstance(headtype, numba.types.Array):
        if advancedtype == NOTADVANCED:
            def getitem(array, head, tail, advanced):
                offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
                offsets[0] = 0
                index = numpy.empty(len(head)*len(array.starts), numpy.int64)
                nextadvanced = numpy.empty(len(index), numpy.int64)

                k = 0
                for i in range(len(array.starts)):
                    length = array.stops[i] - array.starts[i]

                    for j in range(len(head)):
                        norm = head[j]
                        if norm < 0:
                            norm += length
                        if norm < 0 or norm >= length:
                            raise IndexError("advanced index is out of bounds in JaggedArray")
                        index[k] = array.starts[i] + norm
                        nextadvanced[k] = j
                        k += 1
                    offsets[i + 1] = k

                starts = offsets[:-1]
                stops = offsets[1:]
                next = _JaggedArray_getitem_next(array.content[index], tail, nextadvanced)
                return _JaggedArray_new(array, starts, stops, next, True)

        else:
            def getitem(array, head, tail, advanced):
                index = numpy.empty(len(array.starts), numpy.int64)
                nextadvanced = numpy.empty(len(index), numpy.int64)

                for i in range(len(advanced)):
                    length = array.stops[i] - array.starts[i]
                    if advanced[i] >= len(head):
                        raise IndexError("advanced index lengths do not match")
                    norm = head[advanced[i]]
                    if norm < 0:
                        norm += length
                    if norm < 0 or norm >= length:
                        raise IndexError("advanced index is out of bounds in JaggedArray")
                    index[i] = array.starts[i] + norm
                    nextadvanced[i] = i

                next = _JaggedArray_getitem_next(array.content[index], tail, nextadvanced)
                return next

    else:
        raise AssertionError(head)

    sig = sig.return_type(arraytype, headtype, tailtype, advancedtype)
    args = (arrayval, headval, tailval, advancedval)
    return context.compile_internal(builder, getitem, sig, args)

@numba.generated_jit(nopython=True)
def _spread_advanced(starts, stops, advanced):
    if isinstance(advanced, numba.types.NoneType):
        return lambda starts, stops, advanced: advanced
    else:
        def impl(starts, stops, advanced):
            counts = stops - starts
            nextadvanced = numpy.empty(counts.sum(), numpy.int64)
            k = 0
            for i in range(len(counts)):
                length = counts[i]
                nextadvanced[k : k + counts[i]] = advanced[i]
                k += length
            return nextadvanced
        return impl

######################################################################## overloading ufuncs

### See numba.typing.npydecl for typing, and then maybe lower as usual?

### @numba.extending.lower_builtin(numpy.add, JaggedArrayType, JaggedArrayType)
### @numba.extending.lower_builtin(numpy.add, JaggedArrayType, numba.types.Array)
### @numba.extending.lower_builtin(numpy.add, numba.types.Array, JaggedArrayType)
### ???






######################################################################## other lowered methods in Numba, including reducers

@numba.typing.templates.infer_getattr
class _JaggedArrayType_type_methods(numba.typing.templates.AttributeTemplate):
    key = JaggedArrayType

    def resolve_reducer(self, arraytype, args, kwargs, endtype):
        if len(args) == 0 and len(kwargs) == 0:
            if isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, JaggedArrayType):
                contenttype = self.resolve_reducer(arraytype.contenttype, args, kwargs, endtype)
                return JaggedArrayType(arraytype.startstype, arraytype.stopstype, contenttype.return_type, special=JaggedArrayNumba)()
            elif isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, numba.types.Array):
                if endtype is None:
                    endtype = arraytype.contenttype.dtype
                return numba.types.Array(endtype, 1, arraytype.contenttype.layout)()

    @numba.typing.templates.bound_function("any")
    def resolve_any(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, JaggedArrayNumba.NUMBA_BOOLTYPE)

    @numba.typing.templates.bound_function("all")
    def resolve_all(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, JaggedArrayNumba.NUMBA_BOOLTYPE)

    @numba.typing.templates.bound_function("count")
    def resolve_count(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, JaggedArrayNumba.NUMBA_INDEXTYPE)

    @numba.typing.templates.bound_function("count_nonzero")
    def resolve_count_nonzero(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, JaggedArrayNumba.NUMBA_INDEXTYPE)

    @numba.typing.templates.bound_function("sum")
    def resolve_sum(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, None)

    @numba.typing.templates.bound_function("prod")
    def resolve_prod(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, None)

    @numba.typing.templates.bound_function("min")
    def resolve_min(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, None)

    @numba.typing.templates.bound_function("max")
    def resolve_max(self, arraytype, args, kwargs):
        return self.resolve_reducer(arraytype, args, kwargs, None)

#     @numba.typing.templates.bound_function("structure1d")
#     def resolve_structure1d(self, arraytype, args, kwargs):
#         if len(args) == 0:
#             return arraytype()
#         elif len(args) == 1 and isinstance(args[0], numba.types.NoneType):
#             return arraytype(args[0])
#         elif len(args) == 1 and isinstance(args[0], numba.types.Integer):
#             return arraytype(args[0])

# @numba.extending.lower_builtin("structure1d", JaggedArrayType)
# def _JaggedArray_lower_structure1d_1(context, builder, sig, args):
#     arraytype, = sig.args
#     arrayval, = args
#     return _JaggedArray_lower_structure1d_3(context, builder, sig.return_type(arraytype, JaggedArrayNumba.NUMBA_INDEXTYPE), (arrayval, context.get_constant(JaggedArrayNumba.NUMBA_INDEXTYPE, -1),))

def _JaggedArray_lower_reduce_descend(which, context, builder, sig, args):
    arraytype, = sig.args
    arrayval, = args
    array = numba.cgutils.create_struct_proxy(arraytype)(context, builder, value=arrayval)
    content = which(context, builder, sig.return_type.contenttype(arraytype.contenttype), (array.content,))
    return _JaggedArray_lower_new(context, builder, sig.return_type(arraytype, arraytype.startstype, arraytype.stopstype, sig.return_type.contenttype, JaggedArrayNumba.NUMBA_BOOLTYPE), (arrayval, array.starts, array.stops, content, array.iscompact))

@numba.extending.lower_builtin("any", JaggedArrayType)
def _JaggedArray_lower_any(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_any, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, numpy.bool_)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = False
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]) and array.content[j] != 0:
                    flatout[i] = True
                    break
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("all", JaggedArrayType)
def _JaggedArray_lower_all(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_all, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, numpy.bool_)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = True
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]) and array.content[j] == 0:
                    flatout[i] = False
                    break
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("count", JaggedArrayType)
def _JaggedArray_lower_count(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_count, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, numpy.int64)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = 0
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]):
                    flatout[i] += 1
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("count_nonzero", JaggedArrayType)
def _JaggedArray_lower_count_nonzero(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_count_nonzero, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, numpy.int64)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = 0
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]) and array.content[j] != 0:
                    flatout[i] += 1
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("sum", JaggedArrayType)
def _JaggedArray_lower_sum(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_sum, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, array.content.dtype)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = 0
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]):
                    flatout[i] += array.content[j]
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("prod", JaggedArrayType)
def _JaggedArray_lower_prod(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_prod, context, builder, sig, args)
    def run(array):
        out = numpy.empty(array.starts.shape, array.content.dtype)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = 1
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]):
                    flatout[i] *= array.content[j]
        return out
    return context.compile_internal(builder, run, sig, args)

@numba.extending.lower_builtin("min", JaggedArrayType)
def _JaggedArray_lower_min(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_min, context, builder, sig, args)
    def run(array, identity):
        out = numpy.empty(array.starts.shape, array.content.dtype)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = identity
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]) and array.content[j] < flatout[i]:
                    flatout[i] = array.content[j]
        return out
    datatype = sig.args[0].contenttype.dtype
    if isinstance(datatype, numba.types.Boolean):
        identity = True
    elif isinstance(datatype, numba.types.Integer):
        identity = datatype.maxval
    else:
        identity = numpy.inf
        datatype = numba.types.float64
    return context.compile_internal(builder, run, sig.return_type(sig.args[0], datatype), (args[0], context.get_constant(datatype, identity)))

@numba.extending.lower_builtin("max", JaggedArrayType)
def _JaggedArray_lower_max(context, builder, sig, args):
    if isinstance(sig.args[0].contenttype, JaggedArrayType):
        return _JaggedArray_lower_reduce_descend(_JaggedArray_lower_max, context, builder, sig, args)
    def run(array, identity):
        out = numpy.empty(array.starts.shape, array.content.dtype)
        flatout = out.reshape(-1)
        flatstarts = array.starts.reshape(-1)
        flatstops = array.stops.reshape(-1)
        for i in range(len(flatstarts)):
            flatout[i] = identity
            for j in range(flatstarts[i], flatstops[i]):
                if not math.isnan(array.content[j]) and array.content[j] > flatout[i]:
                    flatout[i] = array.content[j]
        return out
    datatype = sig.args[0].contenttype.dtype
    if isinstance(datatype, numba.types.Boolean):
        identity = False
    elif isinstance(datatype, numba.types.Integer):
        identity = datatype.minval
    else:
        identity = -numpy.inf
        datatype = numba.types.float64
    return context.compile_internal(builder, run, sig.return_type(sig.args[0], datatype), (args[0], context.get_constant(datatype, identity)))

@numba.extending.overload_attribute(JaggedArrayType, "offsets")
def _JaggedArray_offsets(arraytype):
    if arraytype.startstype.ndim == 1:
        def impl(array):
            offsets = numpy.empty(len(array.starts) + 1, numpy.int64)
            if len(array.starts) == 0:
                offsets[0] = 0
                return offsets
            offsets = array.starts[0]
            for i in range(1, len(array.starts)):
                if array.starts[i + 1] != array.stops[i]:
                    raise ValueError("starts and stops are not compatible with a single offsets array; call jagged.compact() first")
                offsets[i] = array.stops[i]
            return offsets
        return impl
    else:
        raise TypeError("len(starts.shape) must be 1 to compute offsets; call jagged.structure1d() first")

@numba.extending.overload_attribute(JaggedArrayType, "counts")
def _JaggedArray_counts(arraytype):
    def impl(array):
        return array.stops - array.starts
    return impl

@numba.extending.overload_attribute(JaggedArrayType, "parents")
def _JaggedArray_parents(arraytype):
    def impl(array):
        out = numpy.full(array.stops.max(), -1, numpy.int64)
        for i in range(len(array.starts)):
            out[array.starts[i]:array.stops[i]] = i
        return out
    return impl

@numba.extending.overload_attribute(JaggedArrayType, "index")
def _JaggedArray_index(arraytype):
    def impl(array):
        out = numpy.full(array.stops.max(), -1, numpy.int64)
        for i in range(len(array.starts)):
            for j in range(array.starts[i], array.stops[i]):
                out[j] = j - array.starts[i]
        return JaggedArray(array.starts, array.stops, out, array.iscompact)

@numba.extending.overload_method(JaggedArrayType, "regular")
def _JaggedArray_regular(arraytype):
    if not isinstance(arraytype.contenttype, numba.types.Array):
        raise TypeError("JaggedArray.content must be a Numpy array to use jagged.regular()")
    def impl(array):
        count = -1
        for i in range(len(array.starts)):
            if count == -1:
                count = array.stops[i] - array.starts[i]
            elif count != array.stops[i] - array.starts[i]:
                raise ValueError("JaggedArray is not regular: different elements have different counts")
        return array.content.reshape(array.starts.shape + (count,))
    return impl

@numba.extending.overload_method(JaggedArrayType, "compact")
def _JaggedArray_compact(arraytype):
    if isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, AwkwardArrayType):
        def impl(array):
            if array.iscompact:
                return array
            if len(array.starts) == 0:
                return _JaggedArray_new(array, array.starts, array.stops[0:0], array.content[0:0], True)

            if array.starts.shape != array.stops.shape:
                raise ValueError("JaggedArray.starts must have the same shape as JaggedArray.stops")
            flatstarts = array.starts.reshape(-1)
            flatstops = array.stops.reshape(-1)

            offsets = numpy.empty(len(flatstarts) + 1, flatstarts.dtype)
            offsets[0] = 0
            for i in range(len(flatstarts)):
                count = flatstops[i] - flatstarts[i]
                if count < 0:
                    raise ValueError("JaggedArray.stops[i] must be greater than or equal to JaggedArray.starts[i] for all i")
                offsets[i + 1] = offsets[i] + count

            index = numpy.empty(offsets[-1], numpy.int64)
            k = 0
            for i in range(len(flatstarts)):
                for j in range(flatstarts[i], flatstops[i]):
                    index[k] = j
                    k += 1

            starts = offsets[:-1].reshape(array.starts.shape)
            stops = offsets[1:].reshape(array.starts.shape)    # intentional
            content = array.content[index]
            return _JaggedArray_new(array, starts, stops, content, True)

        return impl

    elif isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, numba.types.Array):
        def impl(array):
            if array.iscompact:
                return array
            if len(array.starts) == 0:
                return _JaggedArray_new(array, array.starts, array.stops[0:0], array.content[0:0], True)

            if array.starts.shape != array.stops.shape:
                raise ValueError("JaggedArray.starts must have the same shape as JaggedArray.stops")
            flatstarts = array.starts.ravel()
            flatstops = array.stops.ravel()

            offsets = numpy.empty(len(flatstarts) + 1, flatstarts.dtype)
            offsets[0] = 0
            for i in range(len(flatstarts)):
                count = flatstops[i] - flatstarts[i]
                if count < 0:
                    raise ValueError("JaggedArray.stops[i] must be greater than or equal to JaggedArray.starts[i] for all i")
                offsets[i + 1] = offsets[i] + count

            content = numpy.empty(offsets[-1], array.content.dtype)
            k = 0
            for i in range(len(flatstarts)):
                for j in range(flatstarts[i], flatstops[i]):
                    content[k] = array.content[j]
                    k += 1

            starts = offsets[:-1].reshape(array.starts.shape)
            stops = offsets[1:].reshape(array.starts.shape)    # intentional
            return _JaggedArray_new(array, starts, stops, content, True)

        return impl

@numba.extending.overload_method(JaggedArrayType, "flatten")
def _JaggedArray_flatten(arraytype):
    if isinstance(arraytype, JaggedArrayType):
        def impl(array):
            if len(array.starts) == 0:
                return array.content[0:0]
            else:
                a = array.compact()
                return a.content[a.starts[0]:a.stops[-1]]
        return impl

@numba.extending.overload_method(JaggedArrayType, "argmin")
def _JaggedArray_argmin(arraytype):
    if isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, AwkwardArrayType):
        def impl(array):
            return _JaggedArray_new(array, array.starts, array.stops, _JaggedArray_argmin(array.content), array.iscompact)
        return impl
    elif isinstance(arraytype, JaggedArrayType):
        def impl(array):
            return _JaggedArray_argminmax(array, True)
        return impl

@numba.extending.overload_method(JaggedArrayType, "argmax")
def _JaggedArray_argmax(arraytype):
    if isinstance(arraytype, JaggedArrayType) and isinstance(arraytype.contenttype, AwkwardArrayType):
        def impl(array):
            return _JaggedArray_new(array, array.starts, array.stops, _JaggedArray_argmax(array.content), array.iscompact)
        return impl
    elif isinstance(arraytype, JaggedArrayType):
        def impl(array):
            return _JaggedArray_argminmax(array, False)
        return impl
    
@numba.njit
def _JaggedArray_argminmax(array, ismin):
    if len(array.content.shape) != 1:
        raise NotImplementedError("content is not one-dimensional")

    flatstarts = array.starts.reshape(-1)
    flatstops = array.stops.reshape(-1)

    offsets = numpy.empty(len(flatstarts) + 1, numpy.int64)
    offsets[0] = 0
    for i in range(len(flatstarts)):
        if flatstarts[i] == flatstops[i]:
            offsets[i + 1] = offsets[i]
        else:
            offsets[i + 1] = offsets[i] + 1

    starts = offsets[:-1].reshape(array.starts.shape)
    stops = offsets[1:].reshape(array.stops.shape)

    output = numpy.empty(offsets[-1], dtype=numpy.int64)

    if ismin:
        k = 0
        for i in range(len(flatstarts)):
            if flatstops[i] != flatstarts[i]:
                best = array.content[flatstarts[i]]
                bestj = 0
                for j in range(flatstarts[i] + 1, flatstops[i]):
                    if array.content[j] < best:
                        best = array.content[j]
                        bestj = j - flatstarts[i]
                output[k] = bestj
                k += 1

    else:
        k = 0
        for i in range(len(flatstarts)):
            if flatstops[i] != flatstarts[i]:
                best = array.content[flatstarts[i]]
                bestj = 0
                for j in range(flatstarts[i] + 1, flatstops[i]):
                    if array.content[j] > best:
                        best = array.content[j]
                        bestj = j - flatstarts[i]
                output[k] = bestj
                k += 1

    return _JaggedArray_new(array, starts, stops, output, True)

@numba.njit
def _JaggedArray_concatenate_njit(arrays, axis):
    return _JaggedArray_concatenate(arrays, axis)

def _JaggedArray_concatenate(arrays, axis):
    pass

@numba.extending.type_callable(_JaggedArray_concatenate)
def _JaggedArray_concatenate(context):
    return None

@numba.extending.lower_builtin(_JaggedArray_concatenate, numba.types.Tuple, numba.types.Integer)
def _JaggedArray_lower_concatenate(context, builder, sig, args):
    raise NotImplementedError



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import awkward.array.masked
from .base import NumbaMethods

class MaskedArrayNumba(NumbaMethods, awkward.array.masked.MaskedArray):
    pass

class BitMaskedArrayNumba(NumbaMethods, awkward.array.masked.BitMaskedArray):
    pass

class IndexedMaskedArrayNumba(NumbaMethods, awkward.array.masked.IndexedMaskedArray):
    pass



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import struct
import unittest

import numpy

from awkward import *
import awkward.type

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_virtual_nbytes(self):
        assert isinstance(VirtualArray(lambda: [1, 2, 3]).nbytes, int)
        assert VirtualArray(lambda: [1, 2, 3], nbytes=12345).nbytes == 12345

    def test_virtual_nocache(self):
        a = VirtualArray(lambda: [1, 2, 3])
        assert not a.ismaterialized
        assert numpy.array_equal(a[:], numpy.array([1, 2, 3]))
        assert a.ismaterialized

        a = VirtualArray(lambda: range(10))
        assert not a.ismaterialized
        assert numpy.array_equal(a[::2], numpy.array([0, 2, 4, 6, 8]))
        assert a.ismaterialized

        a = VirtualArray(lambda: range(10))
        assert not a.ismaterialized
        assert numpy.array_equal(a[[5, 3, 6, 0, 6]], numpy.array([5, 3, 6, 0, 6]))
        assert a.ismaterialized

        a = VirtualArray(lambda: range(10))
        assert not a.ismaterialized
        assert numpy.array_equal(a[[True, False, True, False, True, False, True, False, True, False]], numpy.array([0, 2, 4, 6, 8]))
        assert a.ismaterialized

    def test_virtual_transientcache(self):
        cache = {}
        a = VirtualArray(lambda: [1, 2, 3], cache=cache)
        assert not a.ismaterialized
        a[:]
        assert a.ismaterialized
        assert list(cache) == [a.TransientKey(id(a))]
        assert list(cache) == [a.key]
        assert numpy.array_equal(cache[a.key], numpy.array([1, 2, 3]))
        del a

    def test_virtual_persistentcache(self):
        cache = {}
        a = VirtualArray(lambda: [1, 2, 3], cache=cache, persistentkey="find-me-again")
        assert not a.ismaterialized
        a[:]
        assert a.ismaterialized
        assert list(cache) == ["find-me-again"]
        assert list(cache) == [a.key]
        assert numpy.array_equal(cache[a.key], numpy.array([1, 2, 3]))
        del a

    def test_virtual_dontmaterialize(self):
        a = VirtualArray(lambda: [1, 2, 3], type=awkward.type.fromnumpy(3, int))
        assert not a.ismaterialized
        assert a.dtype == numpy.dtype(int)
        assert a.shape == (3,)
        assert len(a) == 3
        assert a._array == None
        assert not a.ismaterialized
        assert numpy.array_equal(a[:], numpy.array([1, 2, 3]))
        assert a.ismaterialized



#!/usr/bin/env python

# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

import unittest

import numpy
import pytest

import awkward
awkward_cpp = pytest.importorskip("awkward.cpp")
from awkward.cpp import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_cpp_unbox(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)
        def test(x):
            return 3.14
        test(a)
        test(a2)

    def test_cpp_box(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)
        def test(x):
            return x
        assert test(a).tolist() == a.tolist()
        assert test(a2).tolist() == a2.tolist()

    def test_cpp_len(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 1], a)
        def test1(x):
            return len(x)
        assert test1(a) == 3
        assert test1(a2) == 2

    def test_cpp_getitem_integer(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)
        def test1(x, i, j):
            return x[i][j]
        assert test1(a, 0, 0) == 1.1
        assert test1(a, 0, 1) == 2.2
        assert test1(a, 0, 2) == 3.3
        assert test1(a, 2, 0) == 4.4
        assert test1(a, 2, 1) == 5.5
        def test2(x, i):
            return x[i]
        assert test2(a, 0).tolist() == [1.1, 2.2, 3.3]
        assert test2(a, 1).tolist() == []
        assert test2(a, 2).tolist() == [4.4, 5.5]
        assert test2(a2, 0).tolist() == [[1.1, 2.2, 3.3], []]
        assert test2(a2, 1).tolist() == []
        assert test2(a2, 2).tolist() == [[4.4, 5.5]]
        assert test2(a2, 0).content.tolist() == a.content.tolist()

    def test_cpp_getitem_slice(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)   # [[[1.1, 2.2, 3.3], []], [], [[4.4, 5.5]]]
        def test1(x, i, j):
            return x[i:j]
        assert test1(a, 0, 2).tolist() == [[1.1, 2.2, 3.3], []]
        assert test1(a, 1, 3).tolist() == [[], [4.4, 5.5]]
        assert test1(a2, 0, 2).tolist() == [[[1.1, 2.2, 3.3], []], []]
        assert test1(a2, 1, 3).tolist() == [[], [[4.4, 5.5]]]

    def test_cpp_getitem_intarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        starts = numpy.array([0, 3, 4])
        stops = numpy.array([3, 3, 6])
        content = numpy.array([1.1, 2.2, 3.3, 999, 4.4, 5.5])
        a2 = JaggedArray(starts, stops, content)
        index = numpy.array([2, 2, 0, 1])
        def test1(x, i):
            return x[i]
        z = test1(a, index)
        assert z.tolist() == [[4.4, 5.5], [4.4, 5.5], [1.1, 2.2, 3.3], []]
        assert z.content.tolist() == [1.1, 2.2, 3.3, 4.4, 5.5]
        z2 = test1(a2, index)
        assert z2.tolist() == [[4.4, 5.5], [4.4, 5.5], [1.1, 2.2, 3.3], []]
        assert z2.content.tolist() == [1.1, 2.2, 3.3, 999, 4.4, 5.5]
        #def test2(x, i):
        #    return x[i].compact()
        #z = test2(a, index)
        #assert z.tolist() == [[4.4, 5.5], [4.4, 5.5], [1.1, 2.2, 3.3], []]
        #ssert z.content.tolist() == [4.4, 5.5, 4.4, 5.5, 1.1, 2.2, 3.3]
        #z2 = test2(a2, index)
        #assert z2.tolist() == [[4.4, 5.5], [4.4, 5.5], [1.1, 2.2, 3.3], []]
        #assert z2.content.tolist() == [4.4, 5.5, 4.4, 5.5, 1.1, 2.2, 3.3]
        #a3 = JaggedArray.fromcounts([2, 0, 1], a)
        #assert test1(a3, index).tolist() == [[[4.4, 5.5]], [[4.4, 5.5]], [[1.1, 2.2, 3.3], []], []]

    def test_cpp_getitem_boolarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        starts = numpy.array([0, 3, 4])
        stops = numpy.array([3, 3, 6])
        content = numpy.array([1.1, 2.2, 3.3, 999, 4.4, 5.5])
        a2 = JaggedArray(starts, stops, content)
        index = numpy.array([False, True, True])
        def test1(x, i):
            return x[i]
        z = test1(a, index)
        assert z.tolist() == [[], [4.4, 5.5]]
        assert z.content.tolist() == [1.1, 2.2, 3.3, 4.4, 5.5]
        z2 = test1(a2, index)
        assert z2.tolist() == [[], [4.4, 5.5]]
        assert z2.content.tolist() == [1.1, 2.2, 3.3, 999, 4.4, 5.5]
        #def test2(x, i):
        #    return x[i].compact()
        #z = test2(a, index)
        #assert z.tolist() == [[], [4.4, 5.5]]
        #assert z.content.tolist() == [4.4, 5.5]
        #z2 = test2(a2, index)
        #assert z2.tolist() == [[], [4.4, 5.5]]
        #assert z2.content.tolist() == [4.4, 5.5]
        #a3 = JaggedArray.fromcounts([2, 0, 1], a)
        #assert test1(a3, index).tolist() == [[], [[4.4, 5.5]]]

    def test_cpp_getitem_tuple_integer(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        def test1(x, i):
            return x[i,]
        assert test1(a, 2).tolist() == [4.4, 5.5]

        a2 = JaggedArray.fromcounts([2, 0, 1], a)  # [[[1.1, 2.2, 3.3], []], [], [[4.4, 5.5]]]
        def test2(x, i, j):
            return x[i, j]
        assert test2(a2, 0, 0).tolist() == [1.1, 2.2, 3.3]

    def test_cpp_getitem_tuple_slice(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)   # [[[1.1, 2.2, 3.3], []], [], [[4.4, 5.5]]]
        def test1(x, i, j):
            return x[i:j,]
        assert test1(a, 0, 2).tolist() == [[1.1, 2.2, 3.3], []]
        assert test1(a, 1, 3).tolist() == [[], [4.4, 5.5]]
        assert test1(a2, 0, 2).tolist() == [[[1.1, 2.2, 3.3], []], []]
        assert test1(a2, 1, 3).tolist() == [[], [[4.4, 5.5]]]

        a3 = JaggedArray.fromcounts([2, 1], a)   # [[[1.1, 2.2, 3.3], []], [[4.4, 5.5]]]
        assert a3[0:2, 0].tolist() == [[1.1, 2.2, 3.3], [4.4, 5.5]]
        def test2(x, i, j, k):
            return x[i:j, k]
        assert test2(a3, 0, 2, 0).tolist() == [[1.1, 2.2, 3.3], [4.4, 5.5]]

    def test_cpp_getitem_tuple_slice_integer(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]])
        a2 = JaggedArray.fromcounts([2, 2], a)
        def test3(x, i, j, k):
            return x[i:j, k]
        assert test3(a, 0, 2, 1).tolist() == [2.2, 5.5]
        assert test3(a2, 0, 2, 1).tolist() == [[4.4, 5.5], [8.8, 9.9]]

    def test_cpp_getitem_tuple_slice_slice(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]])
        a2 = JaggedArray.fromcounts([2, 2], a)
        def test3(x, i, j, k, l):
            return x[i:j, k:l]
        assert test3(a, 0, 2, -2, None).tolist() == [[2.2, 3.3], [4.4, 5.5]]
        assert test3(a2, 0, 2, 1, 2).tolist() == [[[4.4, 5.5]], [[8.8, 9.9]]]

        def test4(x, i, j, k, l):
            return x[i:j, k:l:1]
        assert test4(a, 0, 2, -2, None).tolist() == [[2.2, 3.3], [4.4, 5.5]]
        assert test4(a2, 0, 2, 1, 2).tolist() == [[[4.4, 5.5]], [[8.8, 9.9]]]

    def test_cpp_getitem_tuple_boolarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]])
        a2 = JaggedArray.fromcounts([2, 2], a)   # [[[1.1, 2.2, 3.3], [4.4, 5.5]], [[6.6, 7.7], [8.8, 9.9]]]
        def test1(x, i, j):
            return x[i, j]
        assert test1(a, numpy.array([True, False, True, False]), 1).tolist() == [2.2, 7.7]
        assert test1(a2, numpy.array([False, True]), 1).tolist() == [[8.8, 9.9]]

    def test_cpp_getitem_tuple_intarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]])
        a2 = JaggedArray.fromcounts([2, 2], a)   # [[[1.1, 2.2, 3.3], [4.4, 5.5]], [[6.6, 7.7], [8.8, 9.9]]]
        def test1(x, i, j):
            return x[i, j]
        assert test1(a, numpy.array([2, 0]), 1).tolist() == [7.7, 2.2]
        assert test1(a2, numpy.array([1, 0, 0]), 1).tolist() == [[8.8, 9.9], [4.4, 5.5], [4.4, 5.5]]

    def test_cpp_getitem_jagged_boolarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)
        def test1(x, i):
            return x[i]
        assert test1(a, JaggedArray.fromiter([[True, False, True], [], [False, True]])).tolist() == [[1.1, 3.3], [], [5.5]]
        assert test1(a2, JaggedArray.fromiter([[True, False], [], [True]])).tolist() == [[[1.1, 2.2, 3.3]], [], [[4.4, 5.5]]]
        assert test1(a2, JaggedArray.fromiter([[[True, False, True], []], [], [[False, True]]])).tolist() == [[[1.1, 3.3], []], [], [[5.5]]]

    def test_cpp_getitem_jagged_intarray(self):
        a = JaggedArray.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        a2 = JaggedArray.fromcounts([2, 0, 1], a)
        def test1(x, i):
            return x[i]
        assert test1(a, JaggedArray.fromiter([[2, 0, 0], [], [1]])).tolist() == [[3.3, 1.1, 1.1], [], [5.5]]
        assert test1(a2, JaggedArray.fromiter([[1, 0], [], [0]])).tolist() == [[[], [1.1, 2.2, 3.3]], [], [[4.4, 5.5]]]
        assert test1(a2, JaggedArray.fromiter([[[2, 0, 0], []], [], [[1]]])).tolist() == [[[3.3, 1.1, 1.1], []], [], [[5.5]]]

    def test_cpp_offsets2parents(self):
        offsets = numpy.array([0, 2, 4, 4, 7], dtype=numpy.int64)
        parents = JaggedArray.offsets2parents(offsets)
        assert parents.tolist() == [0, 0, 1, 1, 3, 3, 3]

    def test_cpp_counts2offsets(self):
        counts = numpy.array([4, 0, 3, 4, 1], dtype=numpy.int64)
        offsets = JaggedArray.counts2offsets(counts)
        assert offsets.tolist() == [0, 4, 4, 7, 11, 12]

    def test_cpp_startsstops2parents(self):
        starts = numpy.array([0, 4, 5, 9], dtype=numpy.int64)
        stops = numpy.array([1, 6, 7, 10], dtype=numpy.int64)
        parents = JaggedArray.startsstops2parents(starts, stops)
        assert parents.tolist() == [0, -1, -1, -1, 1, 2, 2, -1, -1, 3]

    def test_cpp_parents2startsstops(self):
        parents = numpy.array([-1, 0, 0, -1, 2, 2, 2, 3], dtype=numpy.int64)
        startsstops = JaggedArray.parents2startsstops(parents)
        starts = startsstops[0]
        stops = startsstops[1]
        assert starts.tolist() == [1, 0, 4, 7] and stops.tolist() == [3, 0, 7, 8]

    def test_cpp_uniques2offsetsparents(self):
        uniques = numpy.array([0, 3, 4, 6, 8, 8, 9], dtype=numpy.int64)
        offsetsparents = JaggedArray.uniques2offsetsparents(uniques)
        offsets = offsetsparents[0]
        parents = offsetsparents[1]
        assert offsets.tolist() == [0, 1, 2, 3, 4, 6, 7] and parents.tolist() == [0, 1, 2, 3, 4, 4, 5]

    def test_cpp_init(self):
        # This depends on all setters working,
        # as well as str and getitem methods
        a = numpy.array([1, 4, 1, 0])
        b = numpy.array([2, 9, 5, 10])
        c = numpy.arange(10, dtype=numpy.int16)
        test = awkward.cpp.JaggedArray(a, b, c)
        assert str(test) == "[[1] [4 5 6 7 8] [1 2 3 4] [0 1 2 3 4 5 6 7 8 9]]"

    def test_cpp_deepcopy(self):
        a = JaggedArray.fromiter([[1, 5, 2], [], [-6.1]])
        b = a.deepcopy()
        assert a.tolist() == b.tolist()
        assert a.__repr__() != b.__repr__()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import types
import numbers
import re
import keyword
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy

import awkward
import awkward.persist
import awkward.type
import awkward.util

from awkward.util import bothmethod

class AwkwardArray(awkward.util.NDArrayOperatorsMixin):
    """
    AwkwardArray: abstract base class
    """

    allow_tonumpy = True
    allow_iter = True
    check_prop_valid = True
    check_whole_valid = True

    @property
    def awkward(self):
        return awkward

    numpy = numpy
    DEFAULTTYPE = numpy.dtype(numpy.float64)
    CHARTYPE    = numpy.dtype(numpy.uint8)
    INDEXTYPE   = numpy.dtype(numpy.int64)
    TAGTYPE     = numpy.dtype(numpy.uint8)
    MASKTYPE    = numpy.dtype(numpy.bool_)
    BITMASKTYPE = numpy.dtype(numpy.uint8)
    BOOLTYPE    = numpy.dtype(numpy.bool_)

    def __init__(self, *args, **kwds):
        raise TypeError("{0} is an abstract base class; do not instantiate".format(type(self)))

    def __round__(self, n=0):
        if n == 0:
            return self.numpy.rint(self)
        else:
            factor = 10**n
            return self.numpy.rint(self * factor) / factor

    def _checktonumpy(self):
        if not self.allow_tonumpy:
            raise RuntimeError("awkward.array.base.AwkwardArray.allow_tonumpy is False; refusing to convert to Numpy")

    def __array__(self, dtype=None):
        self._checktonumpy()

        if dtype is None:
            dtype = self.dtype

        out = self.numpy.empty(len(self), dtype=dtype)
        for i, x in enumerate(self):
            out[i] = x
        return out

    def __getstate__(self):
        state = {}
        awkward.persist.serialize(self, state)
        return state

    def __setstate__(self, state):
        out = awkward.persist.deserialize(state)
        self.__dict__.update(out.__dict__)
        self.__class__ = out.__class__

    def __reduce__(self):
        state = {}
        awkward.persist.serialize(self, state)
        return (awkward.persist.deserialize, (state,))

    def _checkiter(self):
        if not self.allow_iter:
            raise RuntimeError("awkward.array.base.AwkwardArray.allow_iter is False; refusing to iterate")

    def __iter__(self, checkiter=True):
        if checkiter:
            self._checkiter()
        for i in range(len(self)):
            yield self[i]

    def __str__(self):
        if len(self) <= 6:
            return "[{0}]".format(" ".join(self._util_arraystr(x) for x in self.__iter__(checkiter=False)))

        else:
            first = self[:3]
            if isinstance(first, AwkwardArray):
                first = first.__iter__(checkiter=False)
            last = self[-3:]
            if isinstance(first, AwkwardArray):
                last = last.__iter__(checkiter=False)

            return "[{0} ... {1}]".format(" ".join(self._util_arraystr(x) for x in first), " ".join(self._util_arraystr(x) for x in last))

    def __repr__(self):
        return "<{0} {1} at 0x{2:012x}>".format(self.__class__.__name__, str(self), id(self))

    @property
    def type(self):
        return awkward.type.ArrayType(len(self), awkward.type._resolve(self._gettype({}), {}))

    @property
    def layout(self):
        return awkward.type.Layout(self)

    @property
    def dtype(self):
        return self.type.dtype

    @property
    def shape(self):
        return self.type.shape

    @property
    def ndim(self):
        return len(self.shape)

    def _try_tolist(self, x):
        try:
            return x.tolist()
        except AttributeError:
            return x

    def __bool__(self):
        raise ValueError("The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()")

    __nonzero__ = __bool__

    @property
    def size(self):
        out = 1
        for x in self.shape:
            out *= x
        return out

    @property
    def nbytes(self):
        return self._getnbytes(set())

    def tolist(self):
        import awkward.array.table
        out = []
        for x in self:
            if isinstance(x, awkward.array.table.Table.Row):
                if x._table.istuple:
                    out.append(tuple(self._try_tolist(x[n]) for n in x._table._contents))
                else:
                    out.append(dict((n, self._try_tolist(x[n])) for n in x._table._contents))
            elif isinstance(x, self.numpy.ma.core.MaskedConstant):
                out.append(None)
            else:
                out.append(self._try_tolist(x))
        return out

    def valid(self, exception=False, message=False):
        try:
            self._valid()
        except Exception as err:
            if exception:
                raise err
            elif message:
                return "{0}: {1}".format(type(err), str(err))
            else:
                return False
        else:
            if message:
                return None
            else:
                return True

    def reduce(self, ufunc, identity):
        return self._reduce(ufunc, identity, None)

    def any(self):
        return self._reduce(self.numpy.bitwise_or, False, self.BOOLTYPE)

    def all(self):
        return self._reduce(self.numpy.bitwise_and, True, self.BOOLTYPE)

    def count(self):
        return self._reduce(None, 0, None)

    def count_nonzero(self):
        return self._reduce(self.numpy.count_nonzero, 0, None)

    def sum(self):
        return self._reduce(self.numpy.add, 0, None)

    def prod(self):
        return self._reduce(self.numpy.multiply, 1, None)

    def min(self):
        return self._reduce(self.numpy.minimum, self.numpy.inf, None)

    def max(self):
        return self._reduce(self.numpy.maximum, -self.numpy.inf, None)

    def moment(self, n, weight=None):
        with self.numpy.errstate(invalid="ignore"):
            if weight is None:
                return self.numpy.true_divide((self**n).sum(), self.count())
            else:
                return self.numpy.true_divide(((self * weight)**n).sum(), (self * 0 + weight).sum())

    def mean(self, weight=None):
        with self.numpy.errstate(invalid="ignore"):
            if weight is None:
                return self.numpy.true_divide(self.sum(), self.count())
            else:
                return self.numpy.true_divide((self * weight).sum(), (self * 0 + weight).sum())

    def var(self, weight=None, ddof=0):
        with self.numpy.errstate(invalid="ignore"):
            if weight is None:
                denom = self.count()
                one = self.numpy.true_divide(self.sum(), denom)
                two = self.numpy.true_divide((self**2).sum(), denom)
            else:
                denom (self * 0 + weight).sum()
                one = self.numpy.true_divide((self * weight).sum(), denom)
                two = self.numpy.true_divide(((self * weight)**2).sum(), denom)
            if ddof != 0:
                return (two - one**2) * denom / (denom - ddof)
            else:
                return two - one**2

    def std(self, weight=None, ddof=0):
        with self.numpy.errstate(invalid="ignore"):
            return self.numpy.sqrt(self.var(weight=weight, ddof=ddof))

    def __getattr__(self, where):
        if where in dir(super(AwkwardArray, self)):
            return super(AwkwardArray, self).__getattribute__(where)
        else:
            if where in self.columns:
                try:
                    return self[where]
                except Exception as err:
                    raise AttributeError("while trying to get column {0}, an exception occurred:\n{1}: {2}".format(repr(where), type(err), str(err)))
            else:
                raise AttributeError("no column named {0}".format(repr(where)))

    def __dir__(self):
        return sorted(set(dir(super(AwkwardArray, self)) + [x for x in self.columns if self._dir_pattern.match(x) and not keyword.iskeyword(x)]))
    _dir_pattern = re.compile(r"^[a-zA-Z_]\w*$")

    @property
    def i0(self):
        return self["0"]

    @property
    def i1(self):
        return self["1"]

    @property
    def i2(self):
        return self["2"]

    @property
    def i3(self):
        return self["3"]

    @property
    def i4(self):
        return self["4"]

    @property
    def i5(self):
        return self["5"]

    @property
    def i6(self):
        return self["6"]

    @property
    def i7(self):
        return self["7"]

    @property
    def i8(self):
        return self["8"]

    @property
    def i9(self):
        return self["9"]

    @property
    def ChunkedArray(self):
        import awkward.array.chunked
        return awkward.array.chunked.ChunkedArray

    @property
    def AppendableArray(self):
        import awkward.array.chunked
        return awkward.array.chunked.AppendableArray

    @property
    def IndexedArray(self):
        import awkward.array.indexed
        return awkward.array.indexed.IndexedArray

    @property
    def SparseArray(self):
        import awkward.array.indexed
        return awkward.array.indexed.SparseArray

    @property
    def JaggedArray(self):
        import awkward.array.jagged
        return awkward.array.jagged.JaggedArray

    @property
    def MaskedArray(self):
        import awkward.array.masked
        return awkward.array.masked.MaskedArray

    @property
    def BitMaskedArray(self):
        import awkward.array.masked
        return awkward.array.masked.BitMaskedArray

    @property
    def IndexedMaskedArray(self):
        import awkward.array.masked
        return awkward.array.masked.IndexedMaskedArray

    @property
    def Methods(self):
        import awkward.array.objects
        return awkward.array.objects.Methods

    @property
    def ObjectArray(self):
        import awkward.array.objects
        return awkward.array.objects.ObjectArray

    @property
    def StringArray(self):
        import awkward.array.objects
        return awkward.array.objects.StringArray

    @property
    def Table(self):
        import awkward.array.table
        return awkward.array.table.Table

    @property
    def UnionArray(self):
        import awkward.array.union
        return awkward.array.union.UnionArray

    @property
    def VirtualArray(self):
        import awkward.array.virtual
        return awkward.array.virtual.VirtualArray

    @classmethod
    def _util_isinteger(cls, x):
        return isinstance(x, (numbers.Integral, cls.numpy.integer)) and not isinstance(x, (bool, cls.numpy.bool_, cls.numpy.bool))

    @classmethod
    def _util_isintegertype(cls, x):
        return issubclass(x, cls.numpy.integer) and not issubclass(x, (cls.numpy.bool_, cls.numpy.bool))

    @classmethod
    def _util_toarray(cls, value, defaultdtype, passthrough=None):
        import awkward.array.base
        if passthrough is None:
            passthrough = (cls.numpy.ndarray, AwkwardArray)
        if isinstance(value, passthrough):
            return value
        else:
            try:
                return cls.numpy.frombuffer(value, dtype=getattr(value, "dtype", defaultdtype)).reshape(getattr(value, "shape", -1))
            except (AttributeError, TypeError):
                if len(value) == 0:
                    return cls.numpy.array(value, dtype=defaultdtype, copy=False)
                else:
                    return cls.numpy.array(value, copy=False)

    @classmethod
    def _util_arraystr_draw(cls, x):
        if isinstance(x, tuple):
            return "(" + ", ".join(cls._util_arraystr_draw(y) for y in x) + ")"
        elif isinstance(x, Iterable):
            if len(x) > 6:
                return "[" + " ".join(cls._util_arraystr_draw(y) for y in x[:3]) + " ... " + " ".join(cls._util_arraystr_draw(y) for y in x[-3:]) + "]"
            else:
                return "[" + " ".join(cls._util_arraystr_draw(y) for y in x) + "]"
        else:
            return repr(x)

    @classmethod
    def _util_arraystr(cls, array):
        if isinstance(array, cls.numpy.ndarray):
            return cls._util_arraystr_draw(array)
        elif isinstance(array, AwkwardArray):
            return str(array).replace("\n", "")
        else:
            return repr(array)

    @classmethod
    def _util_isnumpy(cls, dtype):
        if isinstance(dtype, cls.numpy.dtype):
            return True
        else:
            return dtype.isnumpy

    @classmethod
    def _util_deepcopy(cls, array):
        if array is None:
            return None
        elif isinstance(array, cls.numpy.ndarray):
            return array.copy()
        else:
            return array.deepcopy()

    @classmethod
    def _util_hasjagged(cls, array):
        return isinstance(array, AwkwardArray) and array._hasjagged()

    @classmethod
    def _util_counts(cls, array):
        if isinstance(array, AwkwardArray):
            return array.counts
        elif len(array.shape) == 1:
            return cls.numpy.full(array.shape[0], -1, dtype=cls.INDEXTYPE)
        else:
            return cls.numpy.full(array.shape[0], array.shape[1], dtype=cls.INDEXTYPE)

    @classmethod
    def _util_boolmask(cls, array, maskedwhen):
        if isinstance(array, AwkwardArray):
            return array.boolmask(maskedwhen=maskedwhen)
        elif isinstance(array, cls.numpy.ma.MaskedArray) and array.mask is not False:
            if maskedwhen:
                return array.mask
            else:
                return ~array.mask
        else:
            if maskedwhen:
                return cls.numpy.zeros(len(array), dtype=cls.MASKTYPE)
            else:
                return cls.numpy.ones(len(array), dtype=cls.MASKTYPE)

    @property
    def ismasked(self):
        return self.boolmask(maskedwhen=True)

    @property
    def isunmasked(self):
        return self.boolmask(maskedwhen=False)

    @classmethod
    def _util_flattentuple(cls, array):
        if isinstance(array, AwkwardArray):
            return array.flattentuple()
        else:
            return array

    @classmethod
    def _util_flatten(cls, array, axis):
        if isinstance(array, AwkwardArray):
            return array.flatten(axis=axis)
        else:
            axis = min(axis, len(array.shape) - 1)
            return array.reshape(array.shape[:axis] + (-1,) + array.shape[axis + 2:])

    @classmethod
    def _util_pad(cls, array, length, maskedwhen, clip, axis):
        if isinstance(array, AwkwardArray):
            return array.pad(length, maskedwhen=maskedwhen, clip=clip, axis=axis)

        elif len(array.shape) == 1:
            raise ValueError("pad cannot be applied to scalars")

        elif length == 0 and clip:
            if isinstance(maskedwhen, cls.numpy.ma.core.MaskedConstant):
                return cls.JaggedArray.fget(None).fromoffsets([0], cls.numpy.ma.array([]))
            else:
                return cls.JaggedArray.fget(None).fromoffsets([0], cls.MaskedArray.fget(None)([], []))

        elif array.shape[1] > length and clip:
            offsets = cls.numpy.arange(0, length*(len(array) + 1), length, dtype=cls.INDEXTYPE)
            content = array[(slice(None), slice(length)) + array.shape[2:]]
            if isinstance(maskedwhen, cls.numpy.ma.core.MaskedConstant):
                return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.numpy.ma.array(content.reshape((-1,) + array.shape[2:])))
            else:
                return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.MaskedArray.fget(None).fromcontent(content.reshape((-1,) + array.shape[2:]), maskedwhen=maskedwhen))

        elif array.shape[1] >= length:
            offsets = cls.numpy.arange(0, array.shape[1]*(len(array) + 1), array.shape[1], dtype=cls.INDEXTYPE)
            if isinstance(maskedwhen, cls.numpy.ma.core.MaskedConstant):
                return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.numpy.ma.array(array.reshape((-1,) + array.shape[2:])))
            else:
                return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.MaskedArray.fget(None).fromcontent(array.reshape((-1,) + array.shape[2:]), maskedwhen=maskedwhen))

        else:
            offsets = cls.numpy.arange(0, length*(len(array) + 1), length, dtype=cls.INDEXTYPE)
            content = cls.numpy.empty(array.shape[:1] + (length,) + array.shape[2:], dtype=array.dtype)
            content[:, :array.shape[1]] = array
            if isinstance(maskedwhen, cls.numpy.ma.core.MaskedConstant):
                mask = cls.numpy.ones(len(array), dtype=cls.MASKTYPE)
                return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.numpy.ma.array(content.reshape((-1,) + array.shape[2:]), mask=mask))
            if maskedwhen:
                mask = cls.numpy.ones((len(array), length), dtype=cls.MASKTYPE)
            else:
                mask = cls.numpy.zeros((len(array), length), dtype=cls.MASKTYPE)
            mask[:, :array.shape[1]] = not maskedwhen
            return cls.JaggedArray.fget(None).fromoffsets(offsets, cls.MaskedArray.fget(None)(mask.reshape((-1,) + array.shape[2:]), content.reshape((-1,) + array.shape[2:]), maskedwhen=maskedwhen))

    @classmethod
    def _util_regular(cls, array):
        if isinstance(array, AwkwardArray):
            return array.regular()
        else:
            return array

    @classmethod
    def _util_reduce(cls, array, ufunc, identity, dtype):
        if isinstance(array, AwkwardArray):
            return array._reduce(ufunc, identity, dtype)

        elif len(array) == 0:
            if dtype is None:
                dtype = array.dtype
            return ufunc.reduce(cls.numpy.full((1,) + array.shape[1:], identity, dtype=dtype), axis=-1)

        else:
            original = array
            if dtype is not None:
                array = cls.numpy.array(array, dtype=dtype, copy=False)
            if issubclass(array.dtype.type, (cls.numpy.floating, cls.numpy.complexfloating)):
                mask = cls.numpy.isnan(array)
                if mask.any():
                    if array is original or not array.flags.owndata:
                        array = array.copy()
                    array[mask] = identity
            return ufunc.reduce(array, axis=None)

    @classmethod
    def _util_concatenate(cls, arrays):
        if all(isinstance(x, cls.numpy.ndarray) for x in arrays):
            return cls.numpy.concatenate(arrays)
        else:
            return arrays[0].concatenate(arrays[1:])

    @bothmethod
    def concatenate(isclassmethod, cls_or_self, arrays, axis=0):
        if len(arrays) < 1:
            raise ValueError("at least one array needed to concatenate")

        if isclassmethod:
            cls = cls_or_self
        else:
            self = cls_or_self
            cls = type(self)
            arrays = (self,) + tuple(arrays)

        def resolve(t):
            for b in t.__bases__:
                if issubclass(t, AwkwardArray):
                    return resolve(b)
            else:
                return t

        if all(type(x) == cls.numpy.ndarray for x in arrays):
            return cls.numpy.concatenate(arrays, axis=axis)

        if not all(resolve(type(x)) == resolve(type(arrays[0])) for x in arrays):
            if axis == 0:
                tags = cls.numpy.concatenate([cls.numpy.full(len(x), i, dtype=cls.TAGTYPE) for i, x in enumerate(arrays)])
                return cls.UnionArray.fget(None).fromtags(tags, arrays)
            else:
                raise NotImplementedError("axis > 0 for different types")

        for x in arrays:
            x.valid()

        if axis == 0:
            return type(arrays[0])._concatenate_axis0(arrays)
        elif axis == 1:
            return type(arrays[0])._concatenate_axis1(arrays)
        else:
            raise NotImplementedError("axis > 1")

    @classmethod
    def _concatenate_axis0(cls, arrays):
        raise NotImplementedError("{0}.concatenate with axis=0 not implemented".format(cls.__name__))

    @classmethod
    def _concatenate_axis1(cls, arrays):
        raise NotImplementedError("{0}.concatenate with axis=1 not implemented".format(cls.__name__))

    @classmethod
    def _util_isstringslice(cls, where):
        if isinstance(where, awkward.util.string):
            return True
        elif isinstance(where, bytes):
            raise TypeError("column selection must be str, not bytes, in Python 3")
        elif isinstance(where, tuple):
            return False
        elif isinstance(where, (cls.numpy.ndarray, AwkwardArray)) and issubclass(where.dtype.type, (numpy.str, numpy.str_)):
            return True
        elif isinstance(where, (cls.numpy.ndarray, AwkwardArray)) and issubclass(where.dtype.type, (numpy.object, numpy.object_)) and not issubclass(where.dtype.type, (numpy.bool, numpy.bool_)):
            return len(where) > 0 and all(isinstance(x, awkward.util.string) for x in where)
        elif isinstance(where, (cls.numpy.ndarray, AwkwardArray)):
            return False
        try:
            assert len(where) > 0 and all(isinstance(x, awkward.util.string) for x in where)
        except (TypeError, AssertionError):
            return False
        else:
            return True

    @classmethod
    def _util_iscomparison(cls, ufunc):
        return (ufunc is cls.numpy.less or
                ufunc is cls.numpy.less_equal or
                ufunc is cls.numpy.equal or
                ufunc is cls.numpy.not_equal or
                ufunc is cls.numpy.greater or
                ufunc is cls.numpy.greater_equal)

    @classmethod
    def _util_fillna(cls, array, value):
        if isinstance(array, cls.numpy.ndarray):
            data = cls.numpy.ma.getdata(array)
            mask = cls.numpy.ma.getmask(array)
            if mask is not cls.numpy.ma.nomask:
                data[mask] = value
            data[cls.numpy.isnan(data)] = value
            return data
        else:
            return array.fillna(value)

    @classmethod
    def _util_columns_descend(cls, array, seen):
        if isinstance(array, cls.numpy.ndarray):
            if array.dtype.fields is None:
                return []
            else:
                return list(array.dtype.fields)
        else:
            return array._util_columns(seen)

    @property
    def columns(self):
        return self._util_columns(set())

    @classmethod
    def _util_rowname_descend(cls, array, seen):
        if isinstance(array, cls.numpy.ndarray):
            if array.dtype.fields is None:
                raise TypeError("not a Table, so there is no rowname")
            else:
                return None
        else:
            return array._util_rowname(seen)

    @property
    def rowname(self):
        return self._util_rowname(set())

    @property
    def istuple(self):
        columns = self.columns
        return self.rowname == "tuple" and columns == [str(x) for x in range(len(columns))]

    def unzip(self):
        return tuple(self[column_name] for column_name in self._util_columns(set()))

class AwkwardArrayWithContent(AwkwardArray):
    """
    AwkwardArrayWithContent: abstract base class
    """

    def __setitem__(self, where, what):
        if isinstance(where, awkward.util.string):
            self._content[where] = what

        elif self._util_isstringslice(where):
            what = what.unzip()
            if len(where) != len(what):
                raise ValueError("number of keys ({0}) does not match number of provided arrays ({1})".format(len(where), len(what)))
            for x, y in zip(where, what):
                self._content[x] = y

        else:
            raise TypeError("invalid index for assigning column to Table: {0}".format(where))

    def __delitem__(self, where):
        if isinstance(where, awkward.util.string):
            del self._content[where]
        elif self._util_isstringslice(where):
            for x in where:
                del self._content[x]
        else:
            raise TypeError("invalid index for removing column from Table: {0}".format(where))

    def _hasjagged(self):
        return self._util_hasjagged(self._content)

    def _util_columns(self, seen):
        if id(self) in seen:
            return []
        seen.add(id(self))
        return self._util_columns_descend(self._content, seen)

    def _util_rowname(self, seen):
        if id(self) in seen:
            raise TypeError("not a Table, so there is no rowname")
        seen.add(id(self))
        return self._util_rowname_descend(self._content, seen)

    def astype(self, dtype):
        return self.copy(content=self._content.astype(dtype))

    def fillna(self, value):
        return self.copy(content=self._util_fillna(self._content, value))



 #!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

from __future__ import absolute_import
import pandas
import pandas.api.extensions

import awkward.array

def delegated_method(method, index, name, *args, **kwargs):
    return pandas.Series

class Delegated(object):
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, "_index")
        name = object.__getattribute__(obj, "_name")
        result = self._get_result(obj)
        return pandas.Series(result, index, name=name)

class DelegatedProperty(Delegated):
    def _get_result(self, obj, type=None):
        return getattr(object.__getattribute__(obj, "_data"), self.name)

class DelegatedMethod(Delegated):
    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, "_index")
        name = object.__getattribute__(bj, "_name")
        method = getattr(object.__getattribute__(obj, "_data"), self.name)
        return delegated_method(method, index, name)

class AwkwardSeriesDtype(pandas.api.extensions.ExtensionDtype):
    name = "awkward"
    type = awkward.array.base.AwkwardArray
    kind = "O"

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from '{}'".format(cls, string))

@pandas.api.extensions.register_series_accessor("awkward")
class AwkwardAccessor(object):
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)

        self._data = pandas_obj.values
        self._index = pandas_obj.index
        self._name = pandas_obj.name

    @staticmethod
    def _validate(obj):
        if not is_awkward_type(obj):
            raise AttributeError("Cannot use 'awkward' accessor on objects of dtype '{}'.".format(obj.dtype))

def is_awkward_type(obj):
    t = getattr(obj, "dtype", obj)
    try:
        return isinstance(t, AwkwardSeriesDtype) or issubclass(t, AwkwardSeriesDtype)
    except Exception:
        return False

class AwkwardSeries(object):
    @property
    def dtype(self):
        return AwkwardSeriesDtype()

    @staticmethod
    def _findclass(cls):
        for base in cls.__bases__:
            if issubclass(base, awkward.array.base.AwkwardArray):
                if not issubclass(base, AwkwardSeries):
                    return base
                else:
                    out = AwkwardSeries._findclass(base)
                    if out is not None:
                        return out
        return None

    def __array__(self, dtype=None):
        cls = self._findclass(type(self))

        if dtype is None:
            dtype = cls.dtype.fget(self)

        return cls.__array__(self, dtype=dtype)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = self._findclass(type(self)).__array_ufunc__(self, ufunc, method, *inputs, **kwargs)
        return out.pandas

    def isna(self):
        return self.numpy.zeros(self.shape, dtype=self.BOOLTYPE)

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls.concatenate(to_concat)

    @property
    def ChunkedArray(self):
        return mixin(self._findclass(type(self)).ChunkedArray.fget(self))

    @property
    def AppendableArray(self):
        return mixin(self._findclass(type(self)).AppendableArray.fget(self))

    @property
    def IndexedArray(self):
        return mixin(self._findclass(type(self)).IndexedArray.fget(self))

    @property
    def SparseArray(self):
        return mixin(self._findclass(type(self)).SparseArray.fget(self))

    @property
    def JaggedArray(self):
        return mixin(self._findclass(type(self)).JaggedArray.fget(self))

    @property
    def MaskedArray(self):
        return mixin(self._findclass(type(self)).MaskedArray.fget(self))

    @property
    def BitMaskedArray(self):
        return mixin(self._findclass(type(self)).BitMaskedArray.fget(self))

    @property
    def IndexedMaskedArray(self):
        return mixin(self._findclass(type(self)).IndexedMaskedArray.fget(self))

    @property
    def ObjectArray(self):
        return mixin(self._findclass(type(self)).ObjectArray.fget(self))

    @property
    def StringArray(self):
        return mixin(self._findclass(type(self)).StringArray.fget(self))

    @property
    def Table(self):
        return mixin(self._findclass(type(self)).Table.fget(self))

    @property
    def UnionArray(self):
        return mixin(self._findclass(type(self)).UnionArray.fget(self))

    @property
    def VirtualArray(self):
        return mixin(self._findclass(type(self)).VirtualArray.fget(self))

def mixin(tpe):
    return type(tpe._topandas_name, (AwkwardSeries, tpe, pandas.api.extensions.ExtensionArray), {})



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import awkward.array.indexed
from .base import NumbaMethods

class IndexedArrayNumba(NumbaMethods, awkward.array.indexed.IndexedArray):
    pass

class SparseArrayNumba(NumbaMethods, awkward.array.indexed.SparseArray):
    pass



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import os.path
import pybind11
import sys
import setuptools
from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("..", "awkward", "version.py")).read(), g)
    return g["__version__"]

#################################################################################
# The following code is copy-pasted from pybind/python_example/setup.py on github:

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

ext_modules = [
    Extension(
        'awkward.cpp.array.array_impl',
        [os.path.join("awkward", "cpp", "array", "array_impl.cpp")],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True)
        ],
        language='c++'
    ),
]

def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True

def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

############################# end copy-pasted code ##############################

setup(name = "awkward-cpp",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      description = "Connect awkward-arrays to C++ using pybind11.",
      long_description = "",
      author = "Charles Escott",
      author_email = "charlescescott@gmail.com",
      maintainer = "Jim Pivarski (IRIS-HEP)",
      maintainer_email = "pivarski@princeton.edu",
      url = "https://github.com/scikit-hep/awkward-array",
      download_url = "https://github.com/scikit-hep/awkward-array/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["awkward==" + get_version(), "pybind11>=2.2"],
      setup_requires = ["pytest-runner"],
      tests_require = ["pytest"],
      ext_modules = ext_modules,
      cmdclass = {"build_ext": BuildExt},
      classifiers = [
          "Development Status :: 1 - Planning",
          # "Development Status :: 2 - Pre-Alpha",
          # "Development Status :: 3 - Alpha",
          # "Development Status :: 4 - Beta",
          # "Development Status :: 5 - Production/Stable",
          # "Development Status :: 6 - Mature",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import distutils.version

import numpy
if distutils.version.LooseVersion(numpy.__version__) < distutils.version.LooseVersion("1.13.1"):
    raise ImportError("Numpy 1.13.1 or later required")

import warnings
warnings.warn(
    """Consider switching from 'awkward' to 'awkward1', since the new interface will become the default later this year (2020).

    pip install -U awkward1

In Python:

    >>> import awkward1 as ak
    >>> new_style_array = ak.from_awkward0(old_style_array)
    >>> old_style_array = ak.to_awkward0(new_style_array)
""",
    DeprecationWarning
)

from awkward.array.base import AwkwardArray
from awkward.array.chunked import ChunkedArray, AppendableArray
from awkward.array.indexed import IndexedArray, SparseArray
from awkward.array.jagged import JaggedArray
from awkward.array.masked import MaskedArray, BitMaskedArray, IndexedMaskedArray
from awkward.array.objects import Methods, ObjectArray, StringArray
from awkward.array.table import Table
from awkward.array.union import UnionArray
from awkward.array.virtual import VirtualArray

def concatenate(arrays, axis=0):
    return AwkwardArray.concatenate(arrays, axis=axis)

from awkward.generate import fromiter

from awkward.persist import serialize, deserialize, save, load, hdf5

from awkward.arrow import toarrow, fromarrow, toparquet, fromparquet
from awkward.util import topandas

# convenient access to the version number
from awkward.version import __version__

__all__ = ["numpy", "AwkwardArray", "ChunkedArray", "AppendableArray", "IndexedArray", "SparseArray", "JaggedArray", "MaskedArray", "BitMaskedArray", "IndexedMaskedArray", "Methods", "ObjectArray", "Table", "UnionArray", "VirtualArray", "StringArray", "fromiter", "serialize", "deserialize", "save", "load", "hdf5", "toarrow", "fromarrow", "toparquet", "fromparquet", "topandas", "__version__"]

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import unittest

import numpy

from awkward import *
from awkward.type import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_issue49(self):
        a = JaggedArray([2], [5], [1, 2, 3, 4, 5])
        m = JaggedArray([0], [3], [False, False, True])
        assert a[m].tolist() == [[5]]

    def test_issue144(self):
        x = fromiter([[], [0.5], [0.6], []])
        isloose = fromiter([[], [False], [True], []])
        assert x[isloose].tolist() == [[], [], [0.6], []]
        assert x.sum().tolist() == [0.0, 0.5, 0.6, 0.0]

    def test_issue163(self):
        a = fromiter([[1, 3], [4, 5]])
        b = a[a.counts > 10]
        assert b[:,:1].tolist() == []

    def test_issue_190(self):
        a = JaggedArray.fromiter([[], []])
        assert a.pad(1).tolist() == [[None], [None]]
        assert a.pad(2).tolist() == [[None, None], [None, None]]
        assert a.pad(3).tolist() == [[None, None, None], [None, None, None]]

    def test_issue_208(self):
        a = awkward.MaskedArray([True, False, False, True, False, True], awkward.fromiter([[1, 2, 3], [4, 5], [6], [7, 8], [10, 11, 12], [999]]))
        assert a.flatten().tolist() == [None, 4, 5, 6, None, 10, 11, 12, None]
        assert (a + 100).flatten().tolist() == [None, 104, 105, 106, None, 110, 111, 112, None]
        a = awkward.MaskedArray([True, False, False, True, False, True], [1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
        assert a.flatten().tolist() == [None, 2.2, 3.3, None, 5.5, None]
        assert (a + 100).flatten().tolist() == [None, 102.2, 103.3, None, 105.5, None]
